import binascii
import logging
import socket

from .base_ethernet import EthernetBaseConnection

logger = logging.getLogger("naludaq.TCP")
_MAX_BUFF_SIZE = 65536


class TCP(EthernetBaseConnection):
    def __init__(self, ip: str, port: int, stop_word: bytes):
        super().__init__()
        self._timeout = 0.01  # FIXME: hardcoded for TRBHM experiment
        self._stop_word = stop_word
        self._address = (ip, port)
        self._socket: socket.socket = None
        self._rx_fifo = bytearray()

    def __del__(self):
        """Closes the connection if one is open."""
        self.close()

    @property
    def address(self) -> tuple[str, int]:
        return self._address

    @property
    def in_waiting(self) -> int:
        """Gets the number of bytes available to read."""
        self._rx_fifo = self.read_all()
        return len(self._rx_fifo)
    
    @property
    def type(self):
        return "tcp"
    
    @property
    def board_ip(self):
        """Gets the IP address of the board. 
        Alias of TCP.address[0] to use the same naming convention as backend.
        """
        return self.address[0]
    
    @property
    def board_port(self):
        """Gets the IP port of the board. 
        Alias of TCP.address[1] to use the same naming convention as backend.
        """
        return self.address[1]

    @property
    def is_open(self) -> bool:
        """Indicates whether the connection is open."""
        is_open = False
        if self._socket is not None:
            is_open = (
                self._socket.fileno() != -1
            )  # fileno() == -1 if connection dropped
        return is_open

    def open(self):
        if self.is_open:
            logger.debug("TCP.open(): connection already established")
            return

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setblocking(True)
        sock.settimeout(self._timeout)
        try:
            sock.connect(self._address)
        except OSError as e:
            sock.close()
            raise ConnectionError("Failed to connect") from e
        self._socket = sock

    def close(self):
        """Close the connection."""
        if self._socket is not None:
            try:
                self._socket.close()
            except:
                pass
            self._socket = None

    def send(self, data):
        """Converts the string to a series of bytes to send in the correct format

        Input:
            data (hex): A hex type string (interface needs 32 bits, so 8 hex chars minimum)
            pause (float): How logn to wait in between send.
        """
        if len(data) % 8 != 0:
            logger.debug(
                "TCP.send(): need multiples of 8 hex chars, you gave me %s", len(data)
            )
            logger.debug("command:%s", data)
            return

        try:
            tosend = binascii.a2b_hex(data)
        except binascii.Error as e:
            raise ValueError("")

        logger.debug("Sending: %s", data)
        try:
            self._socket.send(tosend)
        except Exception as e:
            logger.error("Could not send data to board: %s", e)

    def read_until(self, stopword: bytes):
        """Pulls data from the connection until a set of characters is found.

        Args:
            stopword (bytes): the sequence of characters that indicate
                when to stop reading.

        Returns:
            A bytearray containing the data, including the trailing stop word.
            The buffer is empty or incomplete if timed out
        """
        buff = bytearray()
        lenstop = len(stopword)
        while buff[-lenstop:] != stopword:
            try:
                next = self.read(1)
                buff.extend(next)
            except Exception as e:
                break
            if len(next) == 0:
                logger.debug(
                    "read_until(): did not end with correct stop word (got %s)",
                    buff[-lenstop:],
                )
                break
        if len(buff) == 0:
            logger.debug("Receive timeout, no data returned")
        return buff

    def receive(self):
        """Reads out an event or answer. Both end with a stopword.

        This will *NOT* work on boards where data and registers have different stopwords.
        """
        return self.read_until(self._stop_word)

    def read(self, amount: int) -> bytes:
        """Read bytes"""
        fifo = self._rx_fifo
        while len(fifo) < amount:
            try:
                data = self._socket.recv(amount - len(fifo))
                fifo.extend(data)
            except Exception as e:
                break
            if len(data) == 0:
                break
        self._rx_fifo = fifo[amount:]
        return fifo[:amount]

    def read_all(self) -> bytes:
        """Read all waiting bytes"""
        fifo = self._rx_fifo
        while True:
            try:
                data = self._socket.recv(_MAX_BUFF_SIZE)
                fifo.extend(data)
            except Exception as e:
                break
            if len(data) == 0:
                break
        self._rx_fifo = bytearray()
        return fifo

    def reset_input_buffer(self):
        """Discards any data currently in the input buffer."""
        self.read_all()

    def reset_output_buffer(self):
        """Does nothing."""



from .autotrigger import get_autotrigger
from .data_collector import get_data_collector
from .pedestals import get_pedestals_controller
from .threshold_scan import get_threshold_scanner
from .waiter import EventWaiter



from collections import defaultdict

from naludaq.board import Board
from naludaq.communication import AnalogRegisters, DigitalRegisters


class UDC16ChannelWriter:
    """Writes values for the analog and digital registers to the given channels
    Makes use of write mask, so not all registers have per channel functionality
    """

    def __init__(self, board: Board):
        self.board = board
        self._num_channels = board.channels
        self._num_chips = board.available_chips
        self._channels_per_chip = self._num_channels // self._num_chips
        self.reset_mask_after_write = True

    def write_analog_register(self, name: str, value: int, channels: list[int] = None):
        """Writes the values to the analog registers for the given channels.

        Args:
            name (str): Name of analog register to write to
            value (int): Value to set the register to
            channels (list[int]): Channels to write to, if not specified,
                defaults to all channels
        """
        if not channels:
            channels = range(self._num_channels)
        self._validate_channel_list_or_raise(channels)
        channel_groups = self._group_channels_by_chips(channels)
        for chip, channels in channel_groups.items():
            self._write_register_chip(AnalogRegisters, name, value, chip, channels)
        if self.reset_mask_after_write:
            self.reset_writemask()

    def reset_writemask(self):
        """Reset the write mask register for all chips.

        The writemask register is restored to 0x1FF (broadcast + all channels),
        regardless of what it was set to before. This is important because
        it ensures the other analog registers don't get written improperly
        elsewhere due to the writemask being wrong.
        """
        DigitalRegisters(self.board).write("writemask_l", 0x1FF)
        DigitalRegisters(self.board).write("writemask_r", 0x1FF)

    def _write_register_chip(
        self,
        registers: AnalogRegisters,
        name: str,
        value: int,
        chip: int,
        channels: list[int],
    ):
        """Writes values to the given register per channels for a chip"""
        self._set_write_mask_for_chip(chip, channels)
        registers(self.board, chip).write(name, value)

    def _set_write_mask_for_chip(self, chip: int, channels: list[int]):
        """Sets the writemask to write to the specified channels of a chip

        Args:
            chip (int): Chip to set write mask
            channels (list[int]): List of channels for the given chip
        """
        wrmsk_l = 0
        wrmsk_r = 0
        for channel in channels:
            wrmsk_l_chan, wrmsk_r_chan = self._get_writemask_from_channel(
                channel % self._channels_per_chip
            )
            wrmsk_l = wrmsk_l | wrmsk_l_chan
            wrmsk_r = wrmsk_r | wrmsk_r_chan

        DigitalRegisters(self.board, chips=chip).write("writemask_l", wrmsk_l)
        DigitalRegisters(self.board, chips=chip).write("writemask_r", wrmsk_r)

    def _get_writemask_from_channel(self, channel: int):
        """
        Returns to the bitmask to be written to writemask digital registers that will enable writing
        to ONLY the channel specified.

        Used to write to 'writemask_l' and 'writemask_r' digital registers.

        Args:
            channel (int): The channel to generate a bitmask for, that will select ONLY that channel.

        Returns:
            hex: Bitmasks (left and right) that is 9 bits long, to set the digital registers to.
        """
        if channel >= 8:
            wrmsk_r = 0x100 | 1 << channel - 8
            wrmsk_l = 0x100
        else:
            wrmsk_r = 0x100
            wrmsk_l = 0x100 | 1 << channel

        return wrmsk_l, wrmsk_r

    def _group_channels_by_chips(self, channels: list) -> dict[int, list[int]]:
        """Groups a list of channels into a dict where the key is the chip,
        the value is a list of channels that correspond to the chip
        ```
        Example::

            {
                0: [0, 4, 12, 15],
                2: [32, 35, 47],
            }
        ```
        """
        groups = defaultdict(lambda: list())
        for channel in channels:
            chip = channel // self._channels_per_chip
            groups[chip].append(channel)
        return groups

    def _validate_channel_list_or_raise(self, channels):
        """Validates a list of channels, and raises an error
        if there's a problem with them.

        Args:
            channels: the object that is supposedly a channels list

        Raises:
            TypeError if the channel list is not a list or
                values are not an int
            ValueError if the values are out of range
        """
        if not isinstance(channels, (list, range)):
            raise TypeError(f"Channels needs to be a list, not a {type(channels)}")
        if len(channels) > self._num_channels:
            raise ValueError(f"Too many channels, max is {self._num_channels}")
        for channel in channels:
            if not isinstance(channel, int):
                raise TypeError(f"Channels can only contain integers")
            if channel >= self._num_channels or channel < 0:
                raise ValueError(f"Channel {channel} is out of bounds")



from naludaq.communication import AnalogRegisters, ControlRegisters

from .default import TriggerController

BOARD_SYSCLK = 78.125e6


class TriggerControllerTrbhm(TriggerController):
    """Trigger controller for TRBHM boards."""

    def __init__(self, board):
        super().__init__(board)

        self._num_chips = board.params.get("num_chips", 2)
        self._num_trig_chans_per_chip = board.params.get("num_trig_chans_per_chip", 4)

    @property
    def trigger_rate(self) -> float:
        """Get/Set the trigger rate in Hz.

        The rate is read as the top 16 bits of a 24-bit counter.
        Because of this, the rate is limited to between 5 and 300_000 Hz.
        The boards clock is running at 78.125 MHz, so the rate is calculated as:
        rate = (78.125 MHz / counter_value in Hz) >> 8.

        4.7 Hz is the minimum rate.
        305175.8 Hz is the maximum rate.

        Args:
            rate (int): The trigger rate in Hz, must be between 5 and 300_000 Hz.

        Returns:
            int: The trigger rate in Hz.
        """
        rate_reg = ControlRegisters(self.board).read("trigger_rate_limit")
        try:
            rate_reg = rate_reg["value"]
        except KeyError:
            return 0
        rate_val = rate_reg << 8
        if rate_val == 0:
            return 0
        rate = BOARD_SYSCLK / rate_val
        return rate

    @trigger_rate.setter
    def trigger_rate(self, rate: int):
        if not isinstance(rate, int):
            raise TypeError("Rate must be an integer.")
        if rate < 5 or rate > 300_000:
            raise ValueError("Rate must be between 5 and 300_000 Hz.")

        rate_reg = int(BOARD_SYSCLK / rate) >> 8

        self._write_control_register("trigger_rate_limit", rate_reg)

    def write_triggers(self):
        """Write the trigger values to the hardware.

        This writes the current set triggers to the hardware.
        By setting the trigger_offsets or the trigger_values
        this function is called automatically and the hardware is updated.
        """
        self._set_trigger_thresholds()
        self._set_wbiases()

        return True

    def _set_trigger_thresholds(self):
        """Sets the trigger values to those in the board object and writes to hardware."""
        trigger_values = self.values
        self._write_analog_registers_per_chip("trigger_threshold_{:02}", trigger_values)

    def _set_wbiases(self):
        """Set the wbias registers based on the board trigger values"""
        wbias = {ch: self.wbias[ch] if x > 0 else 0 for ch, x in self.values.items()}
        self._write_analog_registers_per_chip("wbias_{:02}", wbias)

    def _write_analog_registers_per_chip(
        self, reg_name_format: str, values: "list | dict"
    ):
        """Write analog registers with a given register name format using the
        values provided. Writes values to both chips individually.

        Args:
            reg_name_format (str): register name format string, must take one integer argument.
            values (list): list of values to write. Should be all 8 values across both chips.
        """
        for chip in range(self.board.available_chips):
            for idx in range(self._num_trig_chans_per_chip):
                name = reg_name_format.format(idx)
                val = values[idx + chip * self._num_trig_chans_per_chip]
                AnalogRegisters(self.board, [chip]).write(name, val)



import abc
import inspect
import itertools
from typing import Dict, List


class SemitonMeta(type):
    """Metaclass for the semiton type. See the `Semiton` class documentation.

    Instances are stored internally in a dictionary `_instances`.
    This holds a dict of {type: instance registry dict}, since subclasses
    of a semiton should be independent of each other. The instance registry
    is a dict of instances matched to the arguments the instance was constructed
    with. All this information is used to check whether the arguments have been
    used for instantiating a particular _type_ before.
    """

    def __init__(cls, name, bases, dict):
        super(SemitonMeta, cls).__init__(name, bases, dict)

        # dict[type, dict[arguments instantiated with, instance]]
        cls._instances: Dict[type, Dict[tuple, object]] = {}

    def __call__(cls, *args, **kwargs):
        """Called when it's time to make a new instance.
        Compares the current arguments with the ones used to instantiate a previous
        object, and only instantiates if there's a difference in the arguments.

        Kind of like a cache for classes.
        """
        # Unlikely that the __init__ is wrapped, but better to handle it anyway
        constructor_args = inspect.getfullargspec(inspect.unwrap(cls.__init__)).args[1:]
        all_args = _get_argument_dict(constructor_args, args, kwargs)
        all_args = tuple(all_args[x] for x in constructor_args)

        # Each class gets its own instance registry to account for inheritance
        instance_registry: dict = cls._instances.setdefault(cls, {})

        # Instances are recorded along with their constructing arguments
        instance = instance_registry.get(all_args, None)

        if instance is None:
            instance = super(SemitonMeta, cls).__call__(*args, **kwargs)
            instance_registry[all_args] = instance
        return instance


class Semiton(metaclass=SemitonMeta):
    """Metaclass for turning classes into semitons.

    A "semiton" is like a singleton. The difference is that a semiton
    maintains a record of arguments passed to the constructor, and only
    allows instantiation when the arguments are different than those
    used previously to instantiate the class. This can be useful
    for keeping instances alive and reusing them when some kind of
    persistent state is needed that depends on the arguments given
    to the constructor.

    For simple types (e.g. int, str) the value is tested, while
    for other types (e.g. Board) the object's id is tested instead.
    **Not all arguments can be tested!** Unhashable types such as
    `list` are not supported as arguments to the semiton class.

    TL;DR -- one combination of arguments = one instance.

    Example:
    ```
    class MyClass(metaclass=SemitonMeta):
        def __init__(self, x):
            print('New instance')

    MyClass(0) # prints "New Instance"
    MyClass(0) # prints nothing
    MyClass(1) # prints "New Instance"
    ```

    Raises:
        TypeError if the decorator is not applied to a class, or does not have
            an `__init__` method that takes at least one non-"self" argument.
    """


class _SemitonABCMeta(abc.ABCMeta, SemitonMeta):
    """Metaclass for compatibility between Semitons and ABC"""


class SemitonABC(metaclass=_SemitonABCMeta):
    """A version of the Semiton that is compatible with abc.ABC."""


def _get_argument_dict(arg_names: List[str], args: tuple, kwargs: dict):
    """Converts *args and **kwargs to a dictionary of named arguments.

    Args:
        arg_names (List[str]): the list of argument names of a function.
        args (tuple): the positional arguments
        kwargs (dict): the keyword arguments

    Returns:
        A dictionary of {arg_name: arg_value}
    """
    named_args = [(arg_names[i], arg) for i, arg in enumerate(args)]
    return dict(itertools.chain(named_args, kwargs.items()))



from dataclasses import dataclass

from naludaq.helpers.exceptions import WrongDeviceError
from naludaq.devices.spi_device import SPIDevice
from naludaq.devices.spi_bus import SPIBus
from naludaq.communication.spi import SPIConnection


@dataclass
class AD5674RCommands:
    """ "
    C3 C2 C1 C0 Description
    0 0 0 0 No operation
    0 0 0 1 Write to Input Register n where n = 1 to 8, depending on the DAC selected from the address bits in Table 11 (dependent on LDAC)
    0 0 1 0 Update DAC Register n with contents of Input Register n
    0 0 1 1 Write to and update DAC Channel n
    0 1 0 0 Power down/power up the DAC
    0 1 0 1 Hardware LDAC mask register
    0 1 1 0 Software reset (power-on reset)
    0 1 1 1 Internal reference setup register
    1 0 0 0 Set up the daisy-chain enable (DCEN) register
    1 0 0 1 Set up the readback register (readback enable)
    1 0 1 0 Update all channels of the input register simultaneously with the input data
    1 0 1 1 Update all channels of the DAC register and input register simultaneously with the input data
    1 1 0 0 Reserved
    … … … …
    1 1 1 1 No operation, daisy-chain mod
    """

    no_op = 0x0
    write_reg = 0x1
    update_dac_reg = 0x2
    write_and_update_dac_chan = 0x3
    toggle_power = 0x4
    LDAC_mask = 0x5
    software_rst = 0x6
    internal_reference_setup_reg = 0x7
    daisy_chain_enable = 0x8
    readback_en = 0x9
    upd_all_input_from_value = 0xA
    upd_all_dac_from_value = 0xB
    no_op_daisy_chain = 0xF


class AD5674R_SPI(SPIDevice):
    def __init__(
        self,
        spi_connection,
        bus_id,
        device_id,
        channelmap: dict[int, int],
        daisy_chained: bool = True,
    ):
        self._connection: SPIConnection = spi_connection
        self.bus_id: int = bus_id
        self.device_id: int = device_id
        super().__init__(daisy_chained)

        self.channelmap: dict[int, int] = channelmap
        self._commands = AD5674RCommands()

    @property
    def connection(self) -> SPIBus:
        device = self._connection[self.bus_id]
        return device

    @property
    def channelmap(self) -> dict[int, int]:
        return self._channelmap

    @channelmap.setter
    def channelmap(self, value: dict[int, int]):
        self._channelmap = value

    @property
    def init_cmd(self) -> str:
        """Generate the initialization command for the device."""
        dc_en = self._daisy_chained
        cmd = AD5674RCommands.daisy_chain_enable
        message = f"{cmd:1x}0000{dc_en:1X}"
        return message.upper()

    @property
    def daisy_chain_en(self) -> bool:
        return self._daisy_chained

    @daisy_chain_en.setter
    def daisy_chain_en(self, enabled: bool):
        if not isinstance(enabled, bool):
            raise TypeError("enabled must be a bool.")
        self._daisy_chained = enabled

    def init(self):
        """Initialize the device.

        This will set the daisy chain enable bit for the device.
        """
        message = self.init_cmd()

        self._write(message)

    def no_op(self) -> str:
        """Get a no-op command for the device.

        Will return a no-op command for the device. If the device is daisy chained
        then the no-op command will be a daisy chain no-op command.
        """
        if self._daisy_chained is False:
            cmd = AD5674RCommands.no_op
        else:
            cmd = AD5674RCommands.no_op_daisy_chain
        addr = 0
        value = 0
        msg = f"{cmd:1x}{addr:1x}{value:03x}0"

        return msg.upper()

    def set_channel(self, channel: int, value: int):
        """Set a channel to a specific value.

        Args:
            channel: ASIC channel
            value: 12-bit digital value.

        Raise:

        """
        try:
            addr = self._channelmap[channel]
        except KeyError:
            raise WrongDeviceError(f"This device doesn't map to channel: {channel}")
        cmd = AD5674RCommands.write_and_update_dac_chan

        message = f"{cmd:01x}{addr:01x}{value:03x}0".upper()
        self._write(message)

    def broadcast(self, value: int):
        """Broadcast a value to all channels.

        Args:
            value: 12-bit digital value.
        """
        cmd = AD5674RCommands.upd_all_dac_from_value
        addr = 0x0

        message = f"{cmd:1x}{addr:1x}{value:03x}0".upper()
        self._write(message)

    def _write(self, message: str):
        """Write a message to the device.

        The write is passed to the connection object which will handle the message.

        Args:
            message: The message to write.
        """
        self._connection.write(message, self.device_id, self.bus_id)



import logging
from typing import List

from naludaq.helpers.exceptions import InvalidBoardModelError

from .i2c_dac import I2CDACController

LOGGER = logging.getLogger("naludaq.ext_dac.ad5675")


class DACControllerAD5675(I2CDACController):
    def __init__(self, board, send_delay=0.3) -> None:
        """Controller for the AD5675 DAC.

        Args:
            board (Board): the board object.
            send_delay (float): the time to wait between sequential I2C
                writes in seconds.
        """
        super().__init__(board, send_delay=send_delay)
        self._write_cmd = 0b0011

        if board.params["ext_dac"]["chip"] != "ad5675":
            raise InvalidBoardModelError(
                "The given board cannot be used with this DAC controller"
            )

    def _get_write_words(self, channel: int, value: int) -> List[str]:
        """Get words to write to set the DAC value for a channel.

        Args:
            channel (int): the channel to set the DAC value for.
            value (int): the DAC value.

        Returns:
            A list of bytes to send over the I2C bus.
        """
        value_bits = f"{value:016b}"
        return [
            f"{self._write_cmd:04b}{self._channel_mapping[channel]:04b}",
            value_bits[:8],
            value_bits[8:],
        ]



"""Trigger controller for the HDSoC series of eval boards.

The HDSoC trigger circuit operates differently than the typical way
found on other boards. The HDSoC allows the reference voltages to be
set arbitrarily, theoretically giving the user a higher level of
control and precision over where the trigger threshold is set.

That being said, the trigger circuit is also highly non-linear,
making it difficult to set manually.
"""
import logging
from enum import Enum

from naludaq.helpers.helper_functions import type_name

from .default import TriggerController

logger = logging.getLogger("naludaq.trigger_controller_hdsoc")


class TriggerGroup(Enum):
    """Used to select grouped channels for setting trigger thresholds and references."""

    LEFT = 0
    RIGHT = 1
    CH0_15 = 0
    CH16_31 = 1


class TriggerControllerHdsoc(TriggerController):
    """Trigger controller for the HDSoC board

    HDSoC has a different division of the Trigger circuit.
    The trigger is set with an 8-bit value but the upper and lower
    reference voltages for the trigger circuit can be set.
    The upper and lower reference voltages are set for 0-15 and 16-31
    """

    def __init__(self, board):
        super().__init__(board)

        self._trig_thresh_regname = "trigger_threshold_{}"
        self._first_right_channel = 16

    @property
    def max_reference(self) -> int:
        """Maximum trigger reference value"""
        max_ref = 2 ** self.params.get("ref_bits", 4) - 1
        return max_ref

    @property
    def max_threshold(self) -> int:
        """Maximum trigger value"""
        max_val = 2 ** self.params.get("val_bits", 8) - 1
        return max_val

    @property
    def references(self) -> dict[str, tuple[int, int]]:
        """Get the current low/high references for the left and right side

        left side is ch 0_15
        right side is ch 16_31
        """
        return {
            "0_15": self.left_references,
            "16_31": self.right_references,
        }

    @references.setter
    def references(self, values: "dict[int | str, tuple[int, int]]"):
        """Set the low/high references for the left and right side"""
        self._validate_reference_dict_or_raise(values)
        self._set_references(values)

    @property
    def low_references(self) -> dict[str, int]:
        """Get the current low references for the left and right side"""
        return {
            "0_15": self._get_analog_register("sub_ref_neg_sb_left_trigger"),
            "16_31": self._get_analog_register("sub_ref_neg_sb_right_trigger"),
        }

    @low_references.setter
    def low_references(self, values: "dict[str | str, int]"):
        """Set the low references for the left and right side"""
        if not isinstance(values, dict):
            raise TypeError(f"Values must be a dict, not {type_name(values)}")
        for side, value in values.items():
            if side in [0, "left", "0_15", TriggerGroup.LEFT]:
                regname = "sub_ref_neg_sb_left_trigger"
            elif side in [1, "right", "16_31", TriggerGroup.RIGHT]:
                regname = "sub_ref_neg_sb_right_trigger"
            else:
                raise ValueError(f"Invalid side: {side}")
            self._write_analog_register(regname, value)

    @property
    def high_references(self) -> dict[str, int]:
        """Get the current high references for the left and right side"""
        return {
            "0_15": self._get_analog_register("sub_ref_pos_sb_left_trigger"),
            "16_31": self._get_analog_register("sub_ref_pos_sb_right_trigger"),
        }

    @high_references.setter
    def high_references(self, values: "dict[int | str, int]"):
        """Set the high references for the left and right side"""
        if not isinstance(values, dict):
            raise TypeError(f"Values must be a dict, not {type_name(values)}")
        for side, value in values.items():
            if side in [0, "left", "0_15", TriggerGroup.LEFT]:
                regname = "sub_ref_pos_sb_left_trigger"
            elif side in [1, "right", "16_31", TriggerGroup.RIGHT]:
                regname = "sub_ref_pos_sb_right_trigger"
            else:
                raise ValueError(f"Invalid side: {side}")
            self._write_analog_register(regname, value)

    @property
    def left_references(self) -> tuple[int, int]:
        """Get the current low/high references for the left side"""
        return (
            self._get_analog_register("sub_ref_neg_sb_left_trigger"),
            self._get_analog_register("sub_ref_pos_sb_left_trigger"),
        )

    @property
    def right_references(self) -> tuple[int, int]:
        """Get the current low/high references for the right side"""
        return (
            self._get_analog_register("sub_ref_neg_sb_right_trigger"),
            self._get_analog_register("sub_ref_pos_sb_right_trigger"),
        )

    @property
    def tsel(self) -> dict[str, bool]:
        """Get the current tsel state for the left and right side"""
        return {
            "0_15": bool(self._get_analog_register("tsel_left")),
            "16_31": bool(self._get_analog_register("tsel_right")),
        }

    @tsel.setter
    def tsel(self, values: dict[str, bool]):
        """Set the tsel state for the left and right side"""
        if not isinstance(values, dict):
            raise TypeError(f"Values must be a dict, not {type_name(values)}")
        if any(not isinstance(v, bool) for v in values.values()):
            raise TypeError("Values must all be bool")
        if any(not isinstance(k, str) for k in values.keys()):
            raise TypeError("Keys must all be str")
        for side, value in values.items():
            if side in ["left", "0_15", TriggerGroup.LEFT]:
                regname = "tsel_left"
            elif side in ["right", "16_31", TriggerGroup.RIGHT]:
                regname = "tsel_right"
            else:
                raise ValueError(f"Invalid side: {side}")
            self._write_analog_register(regname, value)

    def _set_references(self, values: dict[str, tuple[int, int]]):
        """Update the reference voltage subrange for one side of the trigger circuit.

        Sets the VSS/VDD reference voltages in order to gain higher precision
        of the trigger thresholds around a region of interest.

        Args:
            side (str): the side of the trigger circuit; left or right.
            lower (int): the lower boundary of the subrange in counts.
            upper (int): the upper boundary of the subrange in counts.
        """
        if not isinstance(values, dict):
            raise TypeError(f"Values must be a dict, not {type_name(values)}")
        for side, (lower, upper) in values.items():
            if side in [0, "0_15", "left", TriggerGroup.CH0_15, TriggerGroup.LEFT]:
                side = "left"
            elif side in [
                1,
                "16_31",
                "right",
                TriggerGroup.CH16_31,
                TriggerGroup.RIGHT,
            ]:
                side = "right"
            else:
                raise ValueError(f"Invalid side: {side}")
            logger.debug(
                "Updating references for %s side: lower=%s, upper=%s",
                side,
                lower,
                upper,
            )
            self._validate_reference_range_or_raise(side, lower, upper)
            self._write_analog_register(f"sub_ref_neg_sb_{side}_trigger", lower)
            self._write_analog_register(f"sub_ref_pos_sb_{side}_trigger", upper)

    def set_trigger_edge(self, side: str, rising: bool = True):
        """Set which signal edge to trigger on.

        Shift between positive going signals (rising) and negative (falling).

        Args:
            rising(bool): If true, trigger on positive going signals, else falling edge.

        Raises:
            TypeError if raises is not a bool.
        """
        self._validate_side_or_raise(side)
        if rising not in [True, False]:
            raise TypeError(f"rising must be a bool, got {rising}")
        logger.debug(
            "Setting %s trigger edge to: %s", side, "rising" if rising else "falling"
        )
        register = {
            "left": "tsgn_left",
            "right": "tsgn_right",
        }[side]
        self._write_analog_register(register, not rising)

    def _set_trigger_thresholds(self, trigger_values: "dict[int, int] | None" = None):
        """Set the trigger values for the individual channels using the values
        stored in the board object.
        """
        trigger_values = trigger_values or self.values
        self._validate_trigger_thresholds_or_raise(trigger_values)
        self._set_tsel_enabled()
        logger.debug("Setting trigger thresholds: %s", trigger_values)
        for channel, threshold_value in trigger_values.items():
            register_name = self._trig_thresh_regname.format(channel)
            self._write_analog_register(register_name, threshold_value)

    def _set_wbiases(self):
        """Enable the biasing system if it's not already enabled"""
        # power the dacs which set Vdd and Vss
        self._write_analog_register("ref_output_bias_left", 0x3E8)
        self._write_analog_register("ref_output_bias_bias_left", 0x3E8)

        self._write_analog_register("ref_output_bias_right", 0x3E8)
        self._write_analog_register("ref_output_bias_bias_right", 0x3E8)

        self._write_analog_register("channel_wbias_source_left", 0x400)
        self._write_analog_register("channel_wbias_source_right", 0x400)

    def _set_tsel_enabled(self):
        """Set whether tsel_{l/r} is enabled based on the trigger values.

        A side is enabled if at least one trigger value is non-zero for that side.
        """
        trigger_values = self.values
        split_pos = self._first_right_channel
        left_val = [val for key, val in trigger_values.items() if key < split_pos]
        right_val = [val for key, val in trigger_values.items() if key >= split_pos]
        en_left = any(t != 0 for t in left_val)
        en_right = any(t != 0 for t in right_val)
        logger.debug("Setting tsel_left: %s", en_left)
        self.tsel = {
            "0_15": en_left,
            "16_31": en_right,
        }

    def _set_trigger_offsets(self):
        """Trigger offsets not implemented yet."""
        pass

    def _validate_trigger_thresholds_or_raise(self, trigger_values: dict[int, int]):
        """Validate the trigger values are within the range of the trigger circuit."""
        max_threshold_value = self.max_threshold
        if not isinstance(trigger_values, (dict)):
            raise TypeError(
                f"Trigger values must be list, not {type_name(trigger_values)}"
            )
        if any(not isinstance(v, int) for v in trigger_values.values()):
            raise TypeError("Trigger values must all be int")
        if any(not 0 <= x <= max_threshold_value for x in trigger_values.values()):
            raise ValueError(
                f"Trigger value must be between 0 and {max_threshold_value}"
            )

    def _validate_side_or_raise(self, side: str):
        VALID_SIDES = [
            0,
            1,
            "left",
            "right",
            "0_15",
            "16_31",
            TriggerGroup.LEFT,
            TriggerGroup.RIGHT,
        ]
        if not isinstance(side, str):
            raise TypeError(f"Side must be str, not {type_name(side)}")
        if side not in VALID_SIDES:
            raise ValueError(f"Side must be one of: {VALID_SIDES}. Got: {side}")

    def _validate_reference_dict_or_raise(self, values: dict):
        """Validate reference dict or raise an error.

        The dict must have the form:
        ```
        {
            'left': [lower, upper],
            'right': [lower, upper],
        }
        ```
        """
        if not isinstance(values, dict):
            raise TypeError(f"Reference values must be a dict, not {type_name(values)}")
        if any(not isinstance(x, (list, tuple)) for x in values.values()):
            raise TypeError("Reference value ranges must be list or tuple")
        if any(len(x) != 2 for x in values.values()):
            raise ValueError(
                "Reference value ranges must each be length 2 (lower & upper)"
            )
        for side, (lower, upper) in values.items():
            self._validate_reference_range_or_raise(side, lower, upper)

    def _validate_reference_range_or_raise(self, side: str, lower: int, upper: int):
        """Validate reference range arguments.

        Args:
            side (str): must be "left" or "right"
            lower (int): must be within bounds
            upper (int): must be within bounds
        """
        self._validate_side_or_raise(side)
        for x in (lower, upper):
            if not isinstance(x, int):
                raise TypeError(
                    f"Reference boundary must be int, not {type_name(side)}"
                )
            if not 0 <= x <= self.max_reference:
                raise ValueError(
                    f"Reference boundary is out of bounds (0-{self.max_reference}). Got: {x}"
                )

    def _get_analog_register(self, register: str) -> int:
        """Get the value of an analog register"""
        return self.board.registers["analog_registers"][register]["value"][0]



from collections import deque

from naludaq.daq.workers.packager import PackagerLight


class DebugPackager(PackagerLight):
    """Similar to PackagerLight, the only difference is that
    it also stores sliced packets of raw data coming from the serial buffer
    into a raw_output_buffer. This allows the user to also collect
    data of the incoming byte stream.
    """

    def __init__(
        self,
        input_buffer: deque,
        output_raw_buffer: deque,
        output_buffer: deque,
        output_answers: deque,
        stop_word: bytes,
        frequency: int,
    ):
        """Prepares a thread (not open) which polls the input buffer and pops all data in it
        into a internal buffer. Depending on frequency, it will slice the internal
        buffer once it reaches a stop word, and store the sliced data into an
        output_buffer for parsing, and output_raw_buffer if defined. If there are answers within
        the data, it will separate it and store it in output_answers.

        Args:
            input_buffer (deque): Incoming stream from serial reader, raw bytes
            output_raw_buffer (deque): Stores raw sliced data if defined
            output_buffer (deque): Stores dict'd raw sliced data + info about pkg
            output_answers (deque): Data recieved intermittent to readout
            stop_word (bytes): Stop word to slice on
            frequency (int): How quick to poll the input buffer and slice
        """
        super().__init__(
            input_buffer, output_buffer, output_answers, stop_word, frequency
        )
        self.output_raw_buffer = output_raw_buffer

    def _slice_buffer(self, internal_buffer):
        """
        Slice the internal buffer at the stopword.

        Takes an internal buffer as a bytearray and cuts it at the stopword.
        The output is a list of packages. Once sliced the last part will be the first part of the next batch.

        Args:
            internal_buffer (bytearray): raw data to slice.

        Returns:
            output: a list of packages
            rest: whatever is after the last stopword.
        """
        output, answers, rest = super()._slice_buffer(internal_buffer)
        if self.output_raw_buffer is not None:
            for package in output:
                self.output_raw_buffer.append({"rawdata": package["rawdata"]})
        return output, answers, rest



import logging
from functools import partial

from naludaq.helpers import type_name

from .peripherals_controller import PeripheralsController

logger = logging.getLogger("naludaq.peripherals_hdsoc")


class AodsocPeripheralsController(PeripheralsController):
    def __init__(self, board):
        """AODSOC peripherals controller.

        Provides a way to read various I2C devices on the board.

        Args:
            board (Board): the board object.
        """
        super().__init__(board)
        self._reader_functions = {
            "current": self.read_current_of_chip,
            "vadjn0": partial(self.read_vadjn, 0),
            "vadjn1": partial(self.read_vadjn, 1),
            "chip_voltage": self.read_chip_voltage,
            "ldo_voltage": self.read_ldo_voltage,
        }

    def read_vadjn(self, chip: int) -> float:
        """Reads a VadjN measurement from the board.

        Args:
            chip (int): chip number, must be 0 or 1

        Returns:
            The measurement in Volts.
        """
        self._validate_chip_or_raise(chip)
        voltage = self._read_ltc2990_voltage(f"vadjn_{chip}")
        logger.debug("VadjN (%s): %s V", chip, voltage)
        return voltage

    def read_vadjp(self, chip: int = 0) -> float:
        """Get the vadjp value. It's not hooked up to the voltage monitor."""
        return 0

    def read_chip_voltage(self) -> float:
        voltage = self._read_ltc2990_voltage("chip_voltage")
        logger.debug("Chip voltage: %s V", voltage)
        return voltage

    def read_ldo_voltage(self) -> float:
        voltage = self._read_ltc2990_voltage("ldo_voltage")
        logger.debug("Chip voltage: %s V", voltage)
        return voltage

    def read_current_of_chip(self) -> float:
        v_ldo = self.read_ldo_voltage()
        v_chip = self.read_chip_voltage()
        return self._convert_voltage_to_current(v_ldo - v_chip)

    @staticmethod
    def _validate_chip_or_raise(chip: int):
        """Checks whether a given chip number is 0 or 1.

        Raises:
            TypeError: if chip is not an int
            ValueError: if chip is not valid
        """
        if not isinstance(chip, int):
            raise TypeError(f'Chip must be an int, not "{type_name(chip)}"')
        if chip not in [0, 1]:
            raise ValueError(f"Chip must be 0 or 1, not {chip}")



"""IO module

Collection of IO functions to support disk IO.
"""
import csv
import logging

import numpy as np

LOGGER = logging.getLogger(__name__)


def get_csv_reader(params: dict):
    """Get the correct CSV reader for the board.

    Args:
        params (dict): board params.
        time_per_channel (bool): export time per channel rather than per row.
        include_timing (bool): include timing information in the CSV file.

    Returns:
        CSVReader: CSV reader object.
    """
    try:
        model = params["model"]
    except Exception:
        model = None
    reader = CSVReader
    if model in ["upac96", "udc16"]:
        reader = UpacCsvReader

    return reader()


class CSVReader:
    """
    CSV reader reads from CSV files.
    The CSV reader technically is two parts, a reader and a parser
    The parser is responsible for parsing the data after the reader has read from disk.

    Attributes:
        time_per_channel (bool): export time per channel rather than per row.
        include_timing (bool): include timing information in the CSV file.
    """

    def import_acq_from_csv(self, filename, channels):
        """
        Reads a csv file that has stored events and stored that data into a
        variant of the acquisition structure, called a kacquisition. (Name is a WIP)
        The difference is that a kacq flattens all events into a single list

        Args:
            filename (str): Filename
            channels (list): Specific channels to parse out of the csv

        """
        if not isinstance(channels, list):
            raise TypeError(f"Expected chans to be list, got {type(channels)}")

        with open(filename) as open_file:
            csv_f = csv.reader(open_file)

            acq = {
                "evt_num": [],
                "windnum": [],
                "data": [],
                "time": [],
            }

            for chan in channels:
                acq["evt_num"].append([])
                acq["windnum"].append([])
                acq["data"].append([])
                acq["time"].append([])

            _ = next(csv_f)
            if _[0] == "acq_num":
                for row in csv_f:
                    for index, chan in enumerate(channels):
                        acq["evt_num"][index].append(int(row[1]))
                        acq["windnum"][index].append(int(row[(chan * 3) + 2]))
                        acq["data"][index].append(int(float(row[(chan * 3) + 3])))
                        acq["time"][index].append(int(row[(chan * 3) + 4]))
            else:
                for row in csv_f:
                    for index, chan in enumerate(channels):
                        acq["evt_num"][index].append(int(row[0]))
                        acq["windnum"][index].append(int(row[(chan * 3) + 1]))
                        acq["data"][index].append(int(float(row[(chan * 3) + 2])))
                        acq["time"][index].append(int(row[(chan * 3) + 3]))

        return acq

    def import_pedestals_from_csv(self, filename: str):
        """Reads pedestal data from a csv. Assumes the csv file is formatted like
        the ouput to io.csv_writer.CSVWriter.export_pedestals_csv

        Args:
            filename (path): Filename for the csv file.

        Output:
            dict in pedestal format
        """
        self._validate_csv_file(
            filename, ["acq_num", "evt_num", "windnum", "time"], ["data_ch"]
        )
        pedestals = []
        with open(filename, "r") as src:
            csv_src = csv.reader(src).__iter__()
            next(csv_src)  # Ignore header
            curr_window = 0
            window_vals = []
            for vals in csv_src:
                window = int(vals[2])
                if window != curr_window:
                    curr_window = window
                    pedestals.append(window_vals)
                    window_vals = []
                window_vals.append([float(val) for val in vals[4:]])

        # Check all same length
        for row in pedestals:
            if len(row) != len(pedestals[0]):
                raise ValueError(
                    "Incorrect CSV file - mismatch between window size in different windows"
                )

        # pedestals is now num_windows x num_samples x num_channels, reformat to num_channels x num_windows x num_samples
        pedestals = np.transpose(np.array(pedestals), (2, 0, 1))
        return {"data": pedestals, "raw_data": pedestals[:, :, :, np.newaxis]}

    def import_thresholdscan_from_csv(
        self, filename: str, num_channels: "int | None" = None
    ):
        """Read the results of a threshold scan from a csv file.
        Assumes the csv file is formatted like the output to io.csv_writer.CSVWriter.export_thresholdscan_to_csv

        Args:
            filename (str): full filepath of of csv file to read from
            num_channels (int or None): Number of channels to consider. If None, consider all channels in file.
                Channels not present in the file will be set to 1.

        Output:
            (threshold_values, trigger_values) where:
                threshold_values: np.array of threshold values in the shape len(trigger_values) x num_channels,
                trigger_values: list of trigger values used to generate threshold values
        """
        self._validate_csv_file(filename, ["Trigger value"], ["ch"])
        trigger_values = []
        threshold_values = []
        with open(filename, "r") as src:
            csv_src = csv.reader(src).__iter__()
            header = next(csv_src)
            channels = [int(s[2:]) for s in header[1:]]
            if num_channels is None:
                num_channels = max(channels) + 1
            for row in csv_src:
                trigger_values.append(int(row[0]))
                threshold_values.append([1] * num_channels)
                for ch, val in zip(channels, row[1:]):
                    try:
                        threshold_values[-1][ch] = float(val)
                    except IndexError:
                        pass  # ch > num_channels
        return np.array(threshold_values), trigger_values

    def _validate_csv_file(self, filename, prepended_headers, per_channel_headers):
        """Reads a csv file and validates that it is formatted as the given headers suggest

        Args:
            filename (str): full filepath of the csv file to validate
            prepended_headers (list[str]): list of headers at beginning
            per_channel_headers (list[str]): list of headers which are repeated per channel. Should not include channel number

        Ex:
            expected header: acq_num, evt_num, windnum, time, data_ch0, data_ch1, ...
            prepended_headers: acq_num, evt_num, windnum, time
            per_channel_headers: data_ch

        Returns:
            whether csv file is valid
        """
        NUMBERS = "0123456789"
        with open(filename, "r") as src:
            csv_src = csv.reader(src)
            csv_iter = iter(csv_src)
            header = next(csv_iter)
            header_iter = iter(header)

            # Check prepended headers
            for header_read, header_need in zip(header_iter, prepended_headers):
                if header_read != header_need:
                    raise ValueError(
                        "Incorrect CSV header - got {} but was expecting {}".format(
                            header_read, header_need
                        )
                    )

            # Check per channel headers
            channels = set()
            for col in header_iter:  # Check header_read subset header_need
                if (
                    col.rstrip(NUMBERS) not in per_channel_headers
                    or col in per_channel_headers
                ):
                    raise ValueError(
                        "Incorrect CSV header - got extraneous channel header {}".format(
                            col
                        )
                    )
                for i in range(len(col) - 1, 0, -1):
                    if col[i - 1] not in NUMBERS:
                        break
                channels.add(col[i:])
            for channel in channels:  # Check header_need subset header_read
                for head in per_channel_headers:
                    if head + channel not in header:
                        raise ValueError(
                            "Incorrect CSV header - did not find header {}".format(
                                head + channel
                            )
                        )

            # Check that row sizes match
            for i, row in enumerate(csv_iter):
                if len(row) != len(header):
                    raise ValueError(
                        "Incorrect CSV format - row {} wrong length ({} vs {})".format(
                            i + 1, len(row), len(header)
                        )
                    )
        return True


class UpacCsvReader(CSVReader):
    pass



"""Controllers change one aspects of the board.

The different controllers change how the board operates.
They are divided based on function.

Controllers are one abstraction step over communication. By manipulating
different registers through the communication layer the controllers
change the desired behaviour.

The controller can't import other controllers, a controller controls an aspect of the board
But doesn't create larger behaviors

DESCRIPTIONS:
=============
Controller:
-----------
Base class for controllers, it accepts one init parameter: board
The board is a property and can be get/set for all controllers.

BoardController:
----------------
Control the overarching state of the board.

"""
from .trigger import get_trigger_controller  # noqa: F401
from .biasing_mode import get_biasing_mode_controller  # noqa: F401
from .board import BoardController, UpacBoardController, get_board_controller  # noqa: F401
from .connection import get_connection_controller  # noqa: F401
from .external_dac import get_dac_controller  # noqa: F401
from .gainstages import get_gainstage_controller  # noqa: F401
from .peripherals import get_peripherals_controller  # noqa: F401
from .readout import get_readout_controller  # noqa: F401
from .clock.si5341_controller import Si5341Controller  # noqa: F401
from .clock import get_clock_controller  # noqa: F401
from .tia import get_tia_controller  # noqa: F401



import time
from typing import Optional

from naludaq.backend.managers.base import Manager

_WRITE_MESSAGE_NIBBLES = 8
_WRITE_MESSAGE_NIBBLES_BYTES = _WRITE_MESSAGE_NIBBLES // 2


class BoardIoManager(Manager):
    def __init__(self, board):
        """Utility for I/O with a board connected to the backend.

        Args:
            context (Context): context used to communicate with the backend.
        """
        super().__init__(board)
        self.bundle_mode = self.board.params.get("usb", {}).get("bundle_mode", False)
        self.tx_pause = self.board.params.get("usb", {}).get("tx_pause", 0.0)

    def write(
        self,
        command: "str | bytes | list[str | bytes]",
        bundle: bool = False,
        pause: Optional[float] = None,
    ):
        """Writes a non-read command to the board.

        Args:
            command (str | bytes): command to send.
        """
        if pause is None:
            pause = self.tx_pause
        if bundle is None:
            bundle = self.bundle_mode
        if isinstance(command, list):
            commands = command
        elif isinstance(command, str):
            if (len(command) > _WRITE_MESSAGE_NIBBLES) and bundle is False:
                commands = [
                    command[i : i + _WRITE_MESSAGE_NIBBLES]
                    for i in range(0, len(command), _WRITE_MESSAGE_NIBBLES)
                ]
            else:
                commands = [command]
        elif isinstance(command, bytes):
            if (len(command) > _WRITE_MESSAGE_NIBBLES_BYTES) and bundle is False:
                commands = [
                    command[i : i + _WRITE_MESSAGE_NIBBLES_BYTES].hex()
                    for i in range(0, len(command), _WRITE_MESSAGE_NIBBLES_BYTES)
                ]
            else:
                commands = [command.hex()]
        self.write_all(commands, pause)

    def write_all(self, commands: list["str | bytes"], pause: Optional[float] = None):
        """Writes a list of non-read commands to the board.

        Args:
            commands (list[str | bytes]): the commands to send.
        """
        if pause is None:
            pause = self.tx_pause
        if len(commands) == 0:
            raise ValueError("Need at least one command")
        commands = [c.hex() if isinstance(c, bytes) else c for c in commands]
        self._send(commands, pause)

    def _send(self, commands: list["str | bytes"], pause: Optional[float] = None):
        if pause is None:
            pause = self.tx_pause
        self.context.client.put(
            "/board/raw",
            json={"packages": commands},
        )
        time.sleep(pause)

    def read(self, command: "str | bytes") -> bytes:
        """Sends a read command to the board and gets the response.

        Args:
            command (str | bytes): read command to send.

        Returns:
            bytes: the response.
        """
        response = self.read_all([command])[0]
        return response

    def read_all(self, commands: "list[str | bytes]") -> list[bytes]:
        """Sends several read commands to the board and retrieves the responses.

        Args:
            commands (list[str | bytes]): commands to send

        Returns:
            list[bytes]: The responses from the board
        """
        if len(commands) == 0:
            raise ValueError("Need at least one command")
        commands = [c.hex() if isinstance(c, bytes) else c for c in commands]
        response = self.context.client.get_json(
            "/board/raw",
            json={"packages": commands},
        )

        return [bytes.fromhex(a) for a in response["packages"]]



"""Timing calibration correcter.

Corrects single and batches of events using a dt array.

"""
import numpy as np


class TimingCorrecter:
    """Correct the timing in an event.

    The timing in the delay line is not perfect and can be calibrated
    The chip clock is divided down 128 times by the DLL in the ASIC, the clock
    is reset at ech 128th window boundary, starting at 0.

    THe correcter takes the correction constants for 128 samples and applies to
    the events `time` field.
    """

    def __init__(self, timingcal):

        self.timingcal = timingcal

    def run(self, event):
        """Correct the time on a single event using the dts

        Args:
            event(dict):

        Returns:
            Timing corrected array
        """
        try:
            shp = event["data"].shape
        except AttributeError:
            shp = (len(event["data"]), len(event["data"][0]))
        corr_times = np.zeros(shp)
        dts = self.timingcal

        for chan, windows in enumerate(event["window_labels"]):
            read_winds = len(windows)

            # if a channel has no data
            if read_winds == 0:
                corr_times[chan] = np.nan
                continue

            # Make sure the tile is larger than the following data
            # Make longer to handle uneven amounts of windows
            time_corrs = np.tile(dts, int(np.round(read_winds / 2)) + 2)

            odd_start_window = self._check_odd_startwin(windows)

            if odd_start_window:
                time_corrs = time_corrs[64:-64]
            else:
                time_corrs = time_corrs[:-128]
            time_corrs = time_corrs[: shp[1]]

            corr_time = np.arange(len(time_corrs)) + time_corrs

            corr_times[chan] = corr_time

        return corr_times

    def _check_odd_startwin(self, windows):
        """Returns True if start window is odd"""
        return windows[0] % 2 == 1

    def batch_correct(self, events):
        """
        input a full acqusition and it will add a corrected_time field to each event

        that uses the dts provided
        """

        for event in events:
            corr_time_dict = {}
            corr_times = self.run(event)
            corr_time_dict["corrected_times"] = corr_times
            event.update(corr_time_dict)



"""
Driver for talking over a FTDI interface

Should be interoperable with the linketh package


Marcus Luck - Nalu Scientific
email: marcus@naluscientific.com
January 2022
"""
import binascii
import logging
import time
from typing import Optional

try:
    import ftd2xx as ftd
    import ftd2xx.defines as ftd_defines
except (FileNotFoundError, OSError):
    raise FileNotFoundError("FTDI driver not installed")
except ImportError:
    raise ImportError("ftd2xx not installed. Please install package.")
import numpy as np

from naludaq.board.connections.base_serial import BaseSerialConnection
from naludaq.helpers import validations
from naludaq.helpers.exceptions import FTDIError
from naludaq.parsers.answer_parser import get_answer_parser
from naludaq.tools.ftdi import index_from_comport, index_from_serial, list_ftdi_devices

logger = logging.getLogger("naludaq.FTDI")

_WRITE_MESSAGE_NIBBLES = 8


class FTDI(BaseSerialConnection):
    """FTDI is a communications class for talking to the boards.

    Uses the ftd2xx Python package, which requires d2xx drivers
    to be properly installed. For information about using d2xx:
    https://ftdichip.com/wp-content/uploads/2020/08/D2XX_Programmers_GuideFT_000071.pdf
    """

    def __init__(self, conn_info: dict):
        """Initializes the FTDI object, but does not open a connection.

        Args:
            conn_info (dict): the connection info dict. Required keys: 'speed', 'model',
                and one of 'serial_number' or 'usb_addr'. 'stop_word' is preferred, and defaults
                to 0xCAFE if not present.

        Raises:
            ConnectionError if the connection info dict is missing the required keys.
        """
        super().__init__()
        self.ser: ftd.FTD2XX = None
        self._read_addr = "AD"
        self._write_addr = "AF"

        self._is_open = False
        self.board_num: int = None
        self._serial_number: str = None
        self._dev_word_len = ftd_defines.BITS_8
        self._dev_stop_bits = ftd_defines.STOP_BITS_1
        self._dev_parity = ftd_defines.PARITY_NONE
        self._rx_timeout = 100  # ms
        self._tx_timeout = 100  # ms
        self._rtscts = False
        self.tx_pause = 0.0
        self.bundle_mode = False

        self._baud = conn_info["speed"]
        self._model = conn_info["model"]
        self.stopword = conn_info.get("stop_word", b"\xca\xfe")  # b'\xfa\xce'
        self.new_firmware = conn_info.get("new_firmware", False)
        self.response_length = 4
        self.parse_answer = get_answer_parser()

        serial_number = conn_info.get("serial_number", None)
        board_num = conn_info.get("board_number", None)
        com_port = conn_info.get("usb_addr", None)

        if serial_number:
            logger.debug("using serial_number %s to connect", serial_number)
            self.serial_number = conn_info["serial_number"]
        elif board_num is not None:
            self.port = conn_info["board_number"]
        elif com_port is not None:
            logger.debug("using usb_addr %s to connect", conn_info["usb_addr"])
            try:
                index = index_from_comport(conn_info["usb_addr"])
            except ConnectionError:
                raise
            self._com_port = com_port
            self.port = index
        else:
            raise ConnectionError("FTDI connection requires the serial num or port num")

    def __del__(self):
        """Closes the connection if one is open."""
        self.close()

    @property
    def baud(self):
        """Get/Set baudrate.

        Must be a positive integer.

        Raises:
            ValueError if set to a value which is not an integer or is non-positive.
        """
        return self._baud

    @baud.setter
    def baud(self, value):
        if isinstance(value, int) and value >= 0:
            self._baud = value
        else:
            raise ValueError("Baudrate must be a positive integer.")
        if self.ser:
            try:
                self.ser.setBaudRate(value)
            except ftd.DeviceError as e:
                logger.error("Failed to set baud rate: {e}")
                raise ConnectionError("Failed to set baud rate")

    @property
    def port(self):
        """Get/set the port. For an FTDI device, the port is the device index).

        Raises:
            TypeError if set to a non-integer type.
            ValueError if set to a device that is not available.
            ConnectionError if set while the connection is open
        """
        return self._port

    @port.setter
    def port(self, value):
        if self._is_open:
            raise ConnectionError("Connection cannot be changed while open.")
        elif not isinstance(value, int):
            raise TypeError("Port number must be a integer.")
        ports = self.get_available_ports()
        logger.debug("available ports:%s", ports)

        if value not in ports:
            raise ValueError(f"{value} is not a valid port.")
        self._port = value

    @property
    def serial_number(self) -> str:
        """Get/set the serial number.

        Sets the device index (port), given part of a serial number.

        Raises:
            TypeError if set to a non-string type.
            ConnectionError if set while the connection is open
        """
        return self._serial_number

    @serial_number.setter
    def serial_number(self, value: str):
        if self._is_open:
            logger.error("did not find port matching SN")
            raise ConnectionError("Connection cannot be changed while open.")
        elif not isinstance(value, str):
            raise TypeError("Serial Number must be a string.")

        try:
            index = index_from_serial(value)
        except FTDIError:
            raise ConnectionError(
                "Could not find a device with a matching serial number."
            )
        logger.debug("found port matching given SN with %s", value)

        self._port = index
        self._serial_number = ftd.getDeviceInfoDetail(index)["serial"].decode()

    @property
    def new_firmware(self):
        """Get/Set if the board uses a new firmware or not.

        If a new firmware is used, the register answers are different.
        """
        return self._new_firmware

    @new_firmware.setter
    def new_firmware(self, value: bool):
        self._new_firmware = value
        self.response_length = 4
        if value is True:
            self.response_length = 8
        self.parse_answer = get_answer_parser(value)

    @property
    def in_waiting(self) -> int:
        """Gets the number of bytes waiting in the input buffer."""
        return self.ser.getQueueStatus()

    def get_available_ports(self):
        """Lists available serial ports.

        Returns:
            A list of available serial ports on the system. Empty if no ports are available.
        """
        return [v["index"] for k, v in list_ftdi_devices().items()]
    
    @property
    def com_port(self):
        """Get/set the COM port. This also sets the index to the appropriate value.

        Raises:
            TypeError if set to a non-integer type.
            ValueError if set to a device that is not available.
            ConnectionError if set while the connection is open
        """
        return self._com_port
    
    @com_port.setter
    def com_port(self, value):
        self._com_port = value
        self.port = index_from_comport(value)
    
    @property
    def type(self):
        return "ftdi"

    @property
    def is_open(self) -> bool:
        """Indicates whether the connection is open."""
        # The `_is_open` attribute is a good start, but is not foolproof (e.g. if device unplugged)
        is_open = self._is_open
        if is_open:
            try:
                _ = self.ser.getQueueStatus()
            except:
                is_open = False
        return is_open

    def open(self):
        """Open connection over the FTDI interface.

        Set board (device) number and baud before opening a connection, use:
            .board_number = value
            .port = value
        """
        serial_connection = self.ser

        board_num = self.port
        try:
            if self.ser:
                if not self._is_open:
                    serial_connection = ftd.open(board_num)
                logger.debug("FTDI.open(): device already open")
            else:
                serial_connection = ftd.open(board_num)
        except ValueError as e_msg:
            logger.error("FTDI.open raise: %s", e_msg)
            serial_connection = None
        except Exception as error_msg:
            logger.debug(error_msg)
            serial_connection = None
        else:
            logger.info("Success!  Connected. %s", board_num)
            try:
                serial_connection.setTimeouts(
                    read=self._rx_timeout, write=self._tx_timeout
                )
                serial_connection.setDataCharacteristics(
                    self._dev_word_len,
                    self._dev_stop_bits,
                    self._dev_parity,
                )
                serial_connection.setBaudRate(self.baud)
                serial_connection.setFlowControl(self._rtscts)
            except ftd.DeviceError as e:
                logger.error(f"Failed to configure connection: {e}")
            self._is_open = True
        self.ser = serial_connection

    def close(self):
        """Close the serial port."""
        if self.ser is not None:
            try:
                self.ser.close()
            except Exception as e:
                logger.debug("Failed to close connection due to: %s", e)
            self.ser = None
        self._is_open = False

    def send(self, data, pause: Optional[float] = None, bundle: Optional[bool] = None):
        """Converts the string to a series of bytes to send in the correct format

        Input:
            data (hex): A hex type string (interface needs 32 bits, so 8 hex chars minimum)
            pause (float): How logn to wait in between send.
        """
        if bundle is None:
            bundle = self.bundle_mode
        if pause is None:
            pause = self.tx_pause

        if len(data) % 8 != 0:
            logger.debug(
                "FTDI.send(): need multiples of 8 hex chars, you gave me %s", len(data)
            )
            logger.debug("command:%s", data)
            return

        logger.debug("Sending: %s in %s mode", data, "bundle" if bundle else "single")
        if (len(data) > _WRITE_MESSAGE_NIBBLES) and bundle is False:
            for i in range(0, len(data), _WRITE_MESSAGE_NIBBLES):
                self._send(data[i : i + _WRITE_MESSAGE_NIBBLES], pause)
        else:
            self._send(data, pause)

    def _send(self, data, pause):
        try:
            tosend = binascii.a2b_hex(data)
        except binascii.Error as e:
            logger.error(f"FTDI.send(): could not convert hex data: {e}")
            return

        try:
            self.ser.write(tosend)
        except AttributeError as error_msg:
            logger.error("Can't send command to board, no connection available.")
        time.sleep(pause)

    def read_until(self, stopword: bytes):
        """Pulls data from the connection until a set of characters is found.

        Args:
            stopword (bytes): the sequence of characters that indicate
                when to stop reading.

        Returns:
            A bytearray containing the data, including the trailing stop word.
            The buffer is empty or incomplete if timed out
        """
        lenstop = len(stopword)
        buff = bytearray()
        while True:
            value = None
            try:
                value = self.read(1)
            except Exception as e:
                logger.debug("Receiving data failed due to: %s", e)
            if value == b"":
                logger.debug("Receive timeout, no data returned")
                break

            buff += value
            if buff[-lenstop:] == stopword:
                break

        logger.debug("Received: %s", buff.hex())

        return buff

    def receive(self):
        """Reads out an event or answer. Both end with a stopword.

        This will *NOT* work on boards where data and registers have different stopwords.
        """
        return self.read_until(self.stopword)

    def read(self, amount):
        """Read bytes"""
        try:
            data = self.ser.read(amount)
        except ftd.ftd2xx.DeviceError as e:
            raise FTDIError(f"FTDI.read failed due to: {e}")
        return data

    def read_all(self):
        """Read all waiting bytes"""
        return self.ser.read(self.ser.getQueueStatus())

    def resync(self, sync_int, sync_addr) -> bool:
        """Resync the board if the input buffer is half full.

        Sometimes you go out of sync, and you need to send single chars till it works
        If the baud speed is set different you can end up with the boards input buffer being partially filled.
        Thus it waits for more characters, try send characters until it accepts an 8 char command again.

        Args:
            sync_int(int): value expected when synced.

        Returns:
            True if board is synced, False if syncing failed.
        """
        try:
            sync_regnum = int(sync_addr)
        except TypeError:
            sync_regnum = int(sync_addr, 16)
        self.reset_input_buffer()
        logger.debug("Syncing board...")
        for i in range(0, 8):
            value = b""
            try:
                value = self.readReg(sync_regnum)
                if value == b"":
                    continue
            except:
                logger.debug("FTDI:resync(): sending single character")
                self.ser.write(bytes(b"F"))

            if value == sync_int:
                logger.debug("synced, was off %s chars", i)
                return True
            else:
                continue

        logger.debug("Couldn't sync")
        return False

    def reset_input_buffer(self, iterations: int = 100):
        """Discards any data currently in the boards output buffer.

        This will empty both the computer input buffer and the hardware
        output buffer.

        However, this will only run 100 times at the most which is ~400kb.

        Args:
            iterations (int): maximum number of times to purge the buffer.
                A cap is necessary since the board may be sending data faster
                than it can be purged.
        """
        validations.validate_positive_int_or_raise(iterations)
        for _ in range(iterations):
            if self.in_waiting == 0:
                break
            self.ser.purge(ftd_defines.PURGE_RX)
            time.sleep(0.001)

    def reset_output_buffer(self):
        """Discards any data currently in the output buffer.

        Discarded data will not be received by the board.
        """
        self.ser.purge(ftd_defines.PURGE_TX)

    @property
    def rtscts(self) -> bool:
        """Get/set whether RTS/CTS flow control is enabled."""
        return self._rtscts == ftd_defines.FLOW_RTS_CTS

    @rtscts.setter
    def rtscts(self, value: bool):
        """Enable or disable RTS/CTS.

        Args:
            value (bool): whether to enable.
        """
        if value:
            self._rtscts = ftd_defines.FLOW_RTS_CTS
        else:
            self._rtscts = ftd_defines.FLOW_NONE
        self.ser.setFlowControl(self._rtscts)

    def set_answer_parser(self, old=True):
        """Set's the answer parser"""
        self.parse_answer = get_answer_parser(old)

    def readReg(self, regNum, timeout=10000) -> int:
        """Attempts to read a register from the board.

        Args:
            regNum (int, str): register address
            timeout (int): number of tries

        Returns:
            The register value.

        Raises:
            ValueError if the response is improperly formatted.
        """
        if self._model in ["siread", "upac32", "upaci", "zdigitizer"]:
            return self.readReg_old(regNum=regNum, timeout=timeout)
        else:
            return self.readReg_new(regNum=regNum, timeout=timeout)

    def readReg_old(self, regNum, timeout=10000):
        """readReg implementation for old firmware"""
        command = f"{self._read_addr}{regNum:02x}0000"
        self.send(command)
        # while self.in_waiting < 4:
        #     timeout -= 1
        #     if timeout < 0:
        #         logger.error("No response, timed out")
        #         return -1
        buff = self.receive()
        logger.debug("readReg() Buff: %s, type: %s", buff, type(buff))
        # struct.unpack(): 'H' is for unsigned short,
        # > is because the stat fifo returns things big-endian for who knows why
        value = 0
        if len(buff) == 4:
            try:
                header = int.from_bytes(buff[0:2], byteorder="big", signed=False)
                read_reg = header & 255  # only 8 lsb
                value = int.from_bytes(buff[2:4], byteorder="big", signed=False)
            except Exception:
                logger.error("couldn't unpack output from register: %s", regNum)
                raise ValueError("readReg couldn't unpack the serial response.")

            if read_reg != regNum:
                logger.error(
                    "readReg didn't read expected register, wanted %s and got %s",
                    regNum,
                    read_reg,
                )
                return -1
        else:
            answer = self.parse_answer(buff)
            value = answer["value"]
        try:
            logger.debug("FTDI.readReg(%s) success: %s", read_reg, value)
        except Exception:
            pass

        return value

    def readReg_new(self, regNum, timeout=10000):
        """
        Reads out the register specified by the integer regNum
        Args:
            regNum (int): The registry number to read
            timeout (int): Amount of retries.
        Returns:
            Hex value of the serialread of two bytes.
        """
        self.response_length

        # Send command
        self._send_readreg_cmd(regNum)

        # Wait for response
        buffer = self.receive()  # self._wait_for_readreg_response(response_length)
        # Parse response
        answer = self.parse_answer(buffer)
        # Make sure responses are valid

        try:
            logger.debug(
                "FTDI.readReg(%s) success: %s %s %s",
                answer["read_reg"],
                answer["value"],
                hex(answer["value"]),
                "{0:b}".format(answer["value"]).zfill(16),
            )
        except Exception:
            pass

        return answer["value"]

    def _send_readreg_cmd(self, regNum):
        """Generates and sends a command to read a register.

        Args:
            regNum (str, int): address as int or hex str.
        """
        cmd_id = np.random.randint(0, 2**16)
        if isinstance(regNum, str):
            regNum = int(regNum, 16)
        read_addr = int(self._read_addr, 16)

        try:
            command = f"{read_addr:x}{regNum:02x}{cmd_id:04x}"
        except Exception as error_msg:
            logger.exception(f"send_readreg_cmd create cmd failed: {error_msg}")
        else:
            self.send(command)

    def _wait_for_readreg_response(self, response_length, timeout: int = 300):
        """Polls the input buffer until there is a valid repsonse.

        Args:
            response_length: number of bytes in response (eg. 4 or 8)
            timeout (int): number of miliseconds before timing out.

        Returns:
            bytes returned.
        """
        buff = b""
        while self.in_waiting < response_length:
            timeout -= 1
            time.sleep(0.001)
            if timeout < 0:
                logger.error("No readReg response, timed out")
                break

        buff = self.read_all()

        return buff


def get_handle(dev_id) -> int:
    """Gets the handle associated with a given device.

    Args:
        dev_id (int): the device index

    Returns:
        The handle as an int
    """
    return ftd.getDeviceInfoDetail(dev_id)["handle"]


def open_device_handle(handle: int):
    """Opens a new FTD2XX object using the given handle

    Args:
        handle (int): the device handle

    Returns:
        The FTD2XX object.
    """
    return ftd.FTD2XX(handle)


def close_all():
    """Closes all FTDI connections opened from this process."""
    for i in range(ftd.createDeviceInfoList()):
        h = get_handle(i)
        try:
            d = open_device_handle(h)
        except ftd.DeviceError:
            pass
        else:
            d.close()



"""Module contains the base class for I2C devices.
"""
import logging
import time
import typing

import numpy as np

from naludaq.communication.i2c_registers import I2CRegisters
from naludaq.helpers.exceptions import I2CError

from .device import Device

logger = logging.getLogger("naludaq.i2c_device")
_bytes_like_types = (bytes, bytearray)
_BytesLike = typing.Union[bytes, bytearray]


class I2CDevice(Device):
    def __init__(self, board, addr: int):
        """Base class for software representation of I2C devices.

        Communication with the physical I2C bus is handled through
        register reads/writes with the FPGA.

        The class is not abstract, as it is possible to communicate
        with any I2C device without defining a subclass for it.

        Args:
            board (board): board object
            addr (int): 7-bit address of the device.

        Raises:
            TypeError if the address is an invalid type.
            ValueError if the address is out of bounds.
        """
        self._validate_addr_or_raise(addr)

        super().__init__()
        self._board = board
        self._addr = addr
        self._send_delay = 0.001  # pause AFTER sending

        self._num_command_regs = board.params.get("i2c", {}).get("max_command_words", 4)
        self._num_response_regs = board.params.get("i2c", {}).get(
            "max_response_words", 4
        )
        self._max_response_width = (self._num_response_regs - 1) * 2

    @property
    def board(self):
        """The board object used to communicate with the device."""
        return self._board

    @property
    def address(self) -> int:
        """The device address (7-bit)"""
        return self._addr

    def write_register(self, reg: int, data: _BytesLike, *, check_ack: bool = False):
        """Writes a register on the device.

        Args:
            reg (int): register to write to
            data (bytes-like): data to write into the register
            check_ack (bool): whether to check if the device acknowledged the operation.

        Raises:
            I2CErrror: if `check_ack=True` and the device did not acknowledge the data sent.
        """
        self._validate_reg_or_raise(reg)
        self._validate_data_or_raise(data)
        self.send_write_command(bytes([reg]) + data, check_ack=check_ack)

    def read_register(self, reg: int, *, width: int, check_ack: bool = False) -> bytes:
        """Reads a register on the device.

        Args:
            reg (int): register to read from
            width (int): width in bytes of the register on the device.
            check_ack (bool): whether to check if the device acknowledged the operation.

        Returns:
            The register value as `bytes`.

        Raises:
            I2CErrror: if `check_ack=True` and the device did not acknowledge the data sent.
        """
        self._validate_reg_or_raise(reg)
        self._validate_reg_width_or_raise(width)

        self.send_write_command(bytes([reg]), check_ack=check_ack)
        time.sleep(0.01)
        self.send_read_command(bytes([0xFF]) * width, check_ack=check_ack)

        response = self._read_response_registers()

        # first byte is junk
        result = self._discard_ack_bits(response)[1 : width + 1]
        return result

    def send_write_command(self, data: _BytesLike, check_ack: bool = False):
        """Sends a write command to the device.

        The rw bit of the address byte is set to 0, indicating a write operation.

        Args:
            data (bytes-like): the data to write to the device, not including the address byte.
            check_ack (bool): whether to check if the device acknowledged the operation.

        Raises:
            I2CErrror: if `check_ack=True` and the device did not acknowledge the data sent.
        """
        self._send_command(data, rw_bit=0, check_ack=check_ack)

    def send_read_command(self, data: _BytesLike, check_ack: bool = False) -> bytes:
        """Sends a read command to the device.

        The rw bit of the address byte is set to 1, indicating a read operation.

        Args:
            data (bytes-like): the data to write to the device, not including the address byte.
            check_ack (bool): whether to check if the device acknowledged the operation.

        Raises:
            I2CErrror: if `check_ack=True` and the device did not acknowledge the data sent.
        """
        self._send_command(data, rw_bit=1, check_ack=check_ack)

    def _send_command(self, data: _BytesLike, *, rw_bit: int, check_ack: bool):
        """Writes to an I2C device.

        Args:
            data (bytes-like): the data to write to the device, not including the address bytes
            rw_bit (int): the r/w bit, 0 (False) for a write, or 1 (True) for a read.
            check_ack (bool): whether to check if the device acknowledged the operation.

        Raises:
            I2CError if the command could not be sent, or if
                `check_ack=True` and the device did not acknowledge the operation.
        """
        self._validate_data_or_raise(data)
        self._validate_check_ack_or_raise(check_ack)
        logger.debug(
            "Writing to device 0x%02x (r/w=%s): %s",
            self._addr,
            rw_bit,
            [hex(x) for x in data],
        )

        addr_byte = (self._addr << 1) | rw_bit
        self._update_fpga_registers(addr_byte, data)
        I2CRegisters(self.board).transmit_command()

        if check_ack:
            time.sleep(0.01)
            response = self._read_response_registers()
            self._validate_acks_or_raise(response, len(data) + 1, rw_bit)

    def _update_fpga_registers(self, addr_byte: int, data: _BytesLike):
        """Updates the I2C registers on the FPGA.

        Changes the following registers:
        - i2c_addr: set to the address of the device, including the r/w bit.
        - i2c_words: set to the number of data bytes
        - i2c_data{0-3}: set to the data _words_. If `len(data) < 8`, the data
        array is left-padded with zeros.

        Args:
            addr_byte (int): the full address byte of the device, including
                the r/w bit.
            words (bytes-like): the bytes to store in the data registers.
        """
        self._write_fpga_register(
            "i2c_words", len(data)
        )  # NOTE: i2c_words is the number of bytes
        self._write_fpga_register("i2c_addr", addr_byte)

        # np.frombuffer() needs an even number of bytes
        if len(data) % 2 == 1:
            data = b"\x00" + data

        words = np.frombuffer(data, ">H")
        words = np.pad(words, (self._num_command_regs - len(words), 0))
        registers = {f"i2c_data{i}": int(value) for i, value in enumerate(words)}
        I2CRegisters(self.board).write_many(registers)

    def _read_response_registers(self) -> "list[int]":
        """Reads all the response registers from the FPGA and splits them into a list.

        Returns:
            list[int]: the responses. Each element is 9 bits wide (8 data + 1 ACK).
        """
        registers = [f"response{i}" for i in range(self._num_response_regs)]
        responses = I2CRegisters(self.board).read_many(registers)
        phrase = "".join(f"{response:016b}" for response in responses)
        phrase = phrase[::-1]
        phrases = [
            int(phrase[i * 9 : (i + 1) * 9], 2)
            for i in range(2 * self._num_response_regs - 1)
        ]
        return phrases

    def _write_fpga_register(self, name: str, value: int):
        """Writes to an I2C register on the FPGA."""
        I2CRegisters(self.board).write(name, value)

    def _read_fpga_register(self, name: str) -> int:
        """Reads an I2C register on the FPGA"""
        return I2CRegisters(self.board).read(name)["value"]

    @staticmethod
    def _validate_acks_or_raise(
        response: "list[int]",
        num_bytes_sent: int,
        rw_bit: int,
    ):
        """Checks whether the ACK/NACK bits in a response follow the
        appropriate sequence of ACK/NACK for a read or write sequence.

        ACK bits are represented by 0, and NACK bits are represented by 1.
        All acknowledge bits should be ACK (0) _except_ for the last bit
        in a read operation.

        Args:
            response (list[int]): the response read from the FPGA registers.
            num_bytes_sent (int): the total number of bytes sent to the device
                in the last read/write operation.
            rw_bit (int): the r/w bit, 1 for read, 0 for write.

        Raises:
            I2CError if the command was not acknowledged by the device.
        """
        # ACK = 0, NACK = 1
        # Last ack bit is NACK for read operations and ACK for write operations
        ack_bits = [x & 1 for x in response[:num_bytes_sent]]
        expected_ack_bits = [0] * (num_bytes_sent - 1) + [rw_bit]
        if ack_bits != expected_ack_bits:
            raise I2CError("Device did not acknowledge the data written")

    @staticmethod
    def _discard_ack_bits(response: "list[int]") -> bytes:
        """Discards the least significant bits (ACK/NACK) in a list of
        response phrases, converting it from a list of 9-bit values
        to bytes.

        Args:
            response (list[int]): the response as a list of 9-bit-wide values.

        Returns:
            bytes: The response without ACK/NACK bits.
        """
        return bytes([x >> 1 for x in response])

    def _validate_reg_width_or_raise(self, width: int):
        """Makes sure the register width is between 0 and the maximum
        allowed for the board.

        Args:
            width (int): the width in bytes

        Raises:
            TypeError: if the width is not an int
            ValueError: if the width is out of bounds
        """
        if not isinstance(width, int):
            raise TypeError("Width must be an int")
        if not 0 < width <= self._max_response_width:
            raise ValueError(
                f'Width "{width}" is out of bounds (1-{self._max_response_width})'
            )

    @staticmethod
    def _validate_addr_or_raise(addr: int):
        """Ensure address is an int and is 7 bits max.

        Unfortunately there's no way of telling whether we've been
        given an 8-bit address which can be represented by fewer, so
        e.g. 8-bit 0x02 (7-bit 0x01) won't raise an error.

        Args:
            addr (int): the address

        Raises:
            TypeError: if the address is not an int
            ValueError: if the address is not in the range 1 - 127 inclusive.
        """
        if not isinstance(addr, int):
            raise TypeError(f"Address must be an int, not {type(addr).__name__}")
        if not 0 <= addr <= 127:
            raise ValueError(f"Address {addr} is out of bounds (0-127)")

    @staticmethod
    def _validate_reg_or_raise(reg: int):
        """Ensure a register number has the right type and value.

        Args:
            reg (int): the register number

        Raises:
            TypeError: if the register number is not an int
            ValueError: if the value is not in the range 0-255 inclusive.
        """
        if not isinstance(reg, int):
            raise TypeError(f"Register must be an int, not {type(reg).__name__}")
        if not 0 <= reg <= 255:
            raise ValueError(f"Register {reg} is out of bounds (0-255)")

    @staticmethod
    def _validate_data_or_raise(data: _BytesLike):
        """Ensure a data array is bytes-like (bytes or bytearray)

        Args:
            data (bytes-like): the data array

        Raises:
            TypeError: if the object has the wrong type.
        """
        if not isinstance(data, _bytes_like_types):
            raise TypeError(
                f"Data must be a bytes-like object, not {type(data).__name__}"
            )

    @staticmethod
    def _validate_check_ack_or_raise(check_ack: bool):
        """Validate the `check_ack` flag.

        Args:
            check_ack (bool): the check_ack flag.

        Raises:
            TypeError: If the flag is not a bool
        """
        if not isinstance(check_ack, bool):
            raise TypeError(
                f"`check_ack` flag must be a bool, not {type(check_ack).__name__}"
            )



from collections import defaultdict
from abc import ABC, abstractmethod


class SPIBus(ABC):
    def __init__(self, board):
        self.board = board
        self._devices = defaultdict(None)
        self._messages = defaultdict(None)
        self._initialized = False

    @abstractmethod
    def init(self):
        self._initialized = True

    def __getitem__(self, index):
        return self._devices[index]

    def __setitem__(self, index, value):
        self._devices[index] = value
        self._messages[index] = None
        self._initialized = False

    def __len__(self):
        return len(self._devices)

    def __iter__(self):
        for dev in self._devices.values():
            yield dev



"""Initializer for the HDSoCv1 board

This initializer will run the init sequence for the HDSoCv1 board.
It contains the HDSoCv1 specific startup sequence.
"""
import logging
import time

from naludaq.board.initializers import Initializers
from naludaq.helpers.decorators import log_function
from naludaq.communication.spi import SPIConnection, SPIBUS_t
from naludaq.devices.ad5674r import AD5674R_SPI

logger = logging.getLogger("naludaq.init_hdsocv2")

VALID_BOARDS = [
    "hdsocv2_eval",
    "hdsocv2_evalr2",
]


# Temporary device mapping, needs to be changed.
# If you see this comment and there are device mappings below, move them to naluconfigs.

DAC_W1_map = {
    0: 12,
    1: 13,
    2: 14,
    3: 15,
    4: 4,
    5: 5,
    6: 6,
    7: 7,
    8: 0,
    9: 1,
    10: 2,
    11: 3,
    12: 8,
    13: 9,
    14: 10,
    15: 11,
}

DAC_W2_map = {
    16: 11,
    17: 10,
    18: 9,
    19: 8,
    20: 3,
    21: 2,
    22: 1,
    23: 0,
    24: 7,
    25: 6,
    26: 5,
    27: 4,
    28: 15,
    29: 14,
    30: 13,
    31: 12,
}

DAC_E1_map = {
    32: 11,
    33: 10,
    34: 9,
    35: 8,
    36: 3,
    37: 2,
    38: 1,
    39: 0,
    40: 7,
    41: 6,
    42: 5,
    43: 4,
    44: 15,
    45: 14,
    46: 13,
    47: 12,
}

DAC_E2_map = {
    48: 12,
    49: 13,
    50: 14,
    51: 15,
    52: 4,
    53: 5,
    54: 6,
    55: 7,
    56: 0,
    57: 1,
    58: 2,
    59: 3,
    60: 8,
    61: 9,
    62: 10,
    63: 11,
}


class InitHDSoCv2(Initializers):
    def __init__(self, board):
        """Initializer for HDSoCv2.

        Args:
            board (Board): the board to initialize.
        """
        super().__init__(board, VALID_BOARDS)
        self.power_sequence = [
            "en_avdd",
            "en_dvdd",
            "en_pllvdd",
            "clk2v5_en",
            "clk1v8_en",
        ]

    def run(self) -> bool:
        """Runs the initialization sequence.

        Returns:
            True, always.
        """
        # Initialize FPGA registers
        self._fpga_init()

        # Power on FMC board
        self._power_toggle(False)
        time.sleep(0.05)
        self._power_toggle(True)
        time.sleep(0.25)

        self._program_clock()

        self._reset_sys()
        self._set_serial_mode(True)
        time.sleep(0.1)
        self.board.control.set_loopback_enabled(False)

        # Chip-side register startup
        self._analog_startup()
        self._digital_startup()

        # init SPI daisy chains
        self._init_spi_connection()

        # Set DACs
        self._init_dacs()

        # Enable ASIC serial output padding  on the FPGA, 12-bit -> 16-bit
        self.control_write("serial_pad_en", True)

        return True

    def _init_spi_connection(self):
        """Initialize the SPI connection for the board."""
        self.board.spi_connection = SPIConnection(self.board)

        self.board.spi_connection.add_bus(
            bus_id=0b01,
            bus_type=SPIBUS_t.DAISY_CHAIN,
        )

        self.board.spi_connection.add_bus(
            bus_id=0b10,
            bus_type=SPIBUS_t.DAISY_CHAIN,
        )

        self.board.spi_connection.add_bus(
            bus_id=0b11,
            bus_type=SPIBUS_t.DAISY_CHAIN,
        )

        self._populate_spi_connection()

        self.board.spi_connection.init()

    def _populate_spi_connection(self):
        """Populate the SPI connection with the devices.

        The devices are connected according to the schematic:
        320-0065-110

        """

        device_map = {
            0: {
                "device_id": 1,
                "bus_id": 0b01,
                "channel_map": DAC_W1_map,
            },
            16: {
                "device_id": 0,
                "bus_id": 0b01,
                "channel_map": DAC_W2_map,
            },
            32: {
                "device_id": 1,
                "bus_id": 0b10,
                "channel_map": DAC_E1_map,
            },
            48: {
                "device_id": 0,
                "bus_id": 0b10,
                "channel_map": DAC_E2_map,
            },
        }

        for channel, values in device_map.items():
            bus_id = values["bus_id"]
            device_id = values["device_id"]
            self.board.spi_connection[bus_id][device_id] = AD5674R_SPI(
                spi_connection=self.board.spi_connection,
                bus_id=values["bus_id"],
                device_id=values["device_id"],
                channelmap=values["channel_map"],
                daisy_chained=True,
            )
            # bus_id=values["bus_id"],
            # device_id=values["device_id"],
            # )

    @log_function(logger)
    def _fpga_init(self):
        """Write all FPGA registers to their default values."""
        self.control_registers.write_all()
        self.i2c_registers.write("i2c_en", True)

    @log_function(logger)
    def _power_toggle(self, state):
        """Toggles the power rails defined in `power_sequence` attribute."""
        for register in self.power_sequence:
            self.control_write(register, state)

    @log_function(logger)
    def _program_clock(self):
        """Program the clock chip, uses the clockfile in the paramters if it exits."""
        self.board.clock.program(filename=self.board.params["clock_file"])
        self.control_write("clk_oeb", False)

    @log_function(logger)
    def _reset_sys(self):
        """Reset the ASIC and clear the analog and digital registers"""
        self.control_write("sysrst", True)
        self.control_write("regclr", True)

        self.control_write("sysrst", False)
        self.control_write("regclr", False)

    @log_function(logger)
    def _set_serial_mode(self, enabled: bool):
        """Turn on or off serial mode."""
        self.board.control.set_serial_mode(enabled)

    @log_function(logger)
    def _digital_startup(self):
        """Startup the digital side of the chip by programming all registers."""
        self.digital_registers.write_all()

    @log_function(logger)
    def _analog_startup(self):
        """Start the analog side of the chip by programming all registers."""
        self.analog_registers.write_all()
        self._dll_startup()

    @log_function(logger)
    def _dll_startup(self):
        """Starting the delay line on the analog side.

        Sets and unsets vanbuff to get the dll going and
        changes the vadjp values to ensure proper SST duty cycle once locked.
        """

        # DLL startup
        for _side in ["left", "right"]:
            self.analog_write(f"qbias_{_side}", 0)
            self.analog_write(f"vadjn_{_side}")  # Use vadjn from yaml
            self.analog_write(f"vanbuf_{_side}", 1000)
            time.sleep(0.5)
            self.analog_write(f"qbias_{_side}", 2950)
            self.analog_write(f"vanbuf_{_side}", 0)

    @log_function(logger)
    def _init_dacs(self):
        """Program the default values to left and right side DACs"""
        channels = list(range(self.board.channels))
        self.board.ext_bias.set_dacs(channels=channels)



import logging
import time

from naludaq.communication.i2c import sendI2cCommand
from naludaq.helpers.exceptions import DACError

from .dac7578 import DACControllerDAC7578

LOGGER = logging.getLogger("naludaq.ext_dac.dac7578_hdsoc")


class DACControllerHDSoC(DACControllerDAC7578):
    def __init__(self, board) -> None:
        """DAC controller for the HDSoCv1, which uses the DAC7578 DAC.

        Args:
            board (Board): the board object.
        """
        super().__init__(board)
        self._power_command = 0b0100
        self._channels_per_device = 8

    def set_dacs(
        self,
        value: "int | None" = None,
        channels: "list[int] | None" = None,
        set_mv: bool = False,
    ):
        """Sets the external DACs for a set of channels.

        Will raise an error if the DACs are disabled.

        Args:
            value (int): value to set the DACs to.
            channels (List[int]): channels to set.
            set_mv (bool): whether the value given is in mV.

        Raises:
            DACError if the DACs are disabled.
        """
        if any(x == "hi-z" for x in self._board.dac_values.values()):
            raise DACError("Cannot set DACs while they are disabled.")
        super().set_dacs(value, channels, set_mv)

    def set_enabled(self, enabled: bool = True):
        """Sets whether the external DAC is enabled for all channels.

        If the external DAC is powered off, the output is set to high impedance
        and the value in `board.dac_values[X]` is set to 'hi-z'.

        On the other hand, when the external DAC is powered on, the value
        is **set to zero.** If you're using this function, take care to
        write the DAC value you desire afterward.

        Args:
            enabled (bool): `True` to power on the DAC, `False` to turn it off
                and set the output to high impedance.

        Raises:
            DACError if the TIA is enabled for any of the given
                channels.
        """
        if not isinstance(enabled, bool):
            raise TypeError('"enabled" must be a boolean')
        if enabled and self._is_tia_enabled():
            raise DACError(
                "Ext. DAC cannot be used for these channels while the TIA is enabled"
            )

        channels = list(range(self._board.channels))
        self._send_power_commands(channels, enabled)
        for channel in channels:
            self.board.dac_values[channel] = 0 if enabled else "hi-z"

        if enabled:
            # Changing the DAC value to 'hi-z' loses the old value;
            # we can at least set to zero for consistent behavior
            self.set_dacs(0, channels)

    def _send_power_commands(self, channels: list[int], enabled: bool):
        """Sends commands to the DAC to enable/disable output for
        the given channels.

        Args:
            channels (List[int]): list of channels to act on
            enabled (bool): whether to enable or disable the given channels
        """
        pd_bits = {
            True: "00",  # powered on
            False: "11",  # powered off & set output to high-Z
        }[enabled]

        # Generate channel states for each device
        # Avoids writing to devices unnecessarily and preserves the state of channels not given
        device_channel_states: dict[
            int, list[str]
        ] = {}  # {device addr: [ch A state, ch B state, ...]}
        for channel in channels:
            device_addr = self._address_mapping[channel]
            bits = device_channel_states.setdefault(
                device_addr, ["0"] * self._channels_per_device
            )
            device_channel = self._channel_mapping[channel]
            bits[device_channel] = "1"

        # Send one command per device
        rw = 0
        for addr, channel_bits in device_channel_states.items():
            LOGGER.debug(
                "Setting DAC enabled: %s for device %s. Channel mask: %s",
                enabled,
                addr,
                {chr(ord("A") + i): v for i, v in enumerate(channel_bits)},
            )

            command = (
                f"{self._power_command:04b}XXXXX{pd_bits}{''.join(channel_bits)}XXXXX"
            )
            command = command.replace("X", "0")
            words = list(map("".join, zip(*[iter(command)] * 8)))  # split into bytes

            sendI2cCommand(self.board, (addr << 1) | rw, words)
            time.sleep(self._send_delay)

    def _is_tia_enabled(self) -> bool:
        """Checks if the TIA is enabled for either side."""
        left = self._board.registers["analog_registers"]["ntia_left"]
        right = self._board.registers["analog_registers"]["ntia_right"]
        if isinstance(left, list):
            left = left[0]
        if isinstance(right, list):
            right = right[0]
        return not left or not right



from abc import ABC, abstractmethod


class BaseConnection(ABC):
    """Base for all connections."""

    def __init__(self):
        pass

    @abstractmethod
    def open(self):
        pass

    @abstractmethod
    def close(self):
        pass

    @abstractmethod
    def send(self):
        pass

    @abstractmethod
    def read(self):
        pass

    @abstractmethod
    def read_all(self):
        pass

    @abstractmethod
    def read_until(self):
        pass

    @abstractmethod
    def receive(self):
        pass

    @abstractmethod
    def reset_input_buffer(self):
        pass

    @abstractmethod
    def reset_output_buffer(self):
        pass

    @property
    @abstractmethod
    def is_open(self) -> bool:
        pass

    @property
    @abstractmethod
    def in_waiting(self):
        pass

    @property
    def type(self):
        return self.__class__.__name__

    def is_uart(self):
        return hasattr(self, "ser")

    def is_ethernet(self):
        return not self.is_uart()



""" Project controller
"""
import gzip
import logging
import os
import pickle
from collections import deque

from naludaq.helpers.exceptions import AcquisitionIOError, ProjectIOError
from naludaq.io import load_pickle_acquisition, save_pickle_acquisition
from naludaq.models import Acquisition

LOGGER = logging.getLogger(__name__)


class ProjectController:
    """ProjectController handles projects, acquisitions and events.

    Project Controller acts as an intermediary between the
    GUI and the Project, Acquisition and events models.
    Every function modifying the datastructure of those objects
    is in this controller.

    Example:
        pc = ProjectController(name, working_directory)

        # Add an acquisition to the project.
        pc.add_acquisition()

        # Get the last acquistion to be added
        acq = pc.get_acquisition()

    """

    def __init__(self, name, working_directory):
        self._project = {
            "name": name,
            "acquisitions": deque(),
            "working_directory": working_directory,
        }
        self._active_acquisition = None
        self._selected_acq = None
        self.maxlen = None

    @property
    def project_name(self):
        """Set/Get the project name.

        Project name is not the same as the project filename.
        The name need to be a str.

        Raises:
            TypeError if the name is not a string.
        """
        return self._project["name"]

    @project_name.setter
    def project_name(self, name):
        if not isinstance(name, (str)):
            raise TypeError("Name must be a string")
        self._project["name"] = str(name)

    @property
    def working_directory(self):
        """The projects working directory.

        Need to be a valid directory.
        Raises:
            NotADirectoryError if the directory is not found.
        """
        return self._project["working_directory"]

    @working_directory.setter
    def working_directory(self, work_dir):
        if not isinstance(work_dir, (str, bytes, os.PathLike)):
            raise TypeError("Working directory must be bytes, string or os.PathLike")
        if not os.path.isdir(work_dir):
            raise NotADirectoryError(f"Directory doesn't exist: {work_dir}")
        self._project["working_directory"] = work_dir

    @property
    def acquisition_notes(self):
        """Get or add notes to the selected acquisition.
        This is notes for the user to store with the selected run.
        Default notes are blank but can be whatever the user needs.

            Returns a blank string if active acquistion is not set.

            Text must be in string format or it'll raise a TypeError.
            Please note the text can contain formatting.
        """
        try:
            return self._active_acquisition.notes
        except AttributeError:
            LOGGER.error("Acquisition notes not found")
            return ""

    @acquisition_notes.setter
    def acquisition_notes(self, text):
        if not isinstance(self._active_acquisition, type(Acquisition())):
            raise NameError("Active acquisition not found, not set yet?")

        try:
            self._active_acquisition.notes = text
        except TypeError as e_msg:
            LOGGER.error(
                "Setting acquisition notes to: %r failed, error %r", text, e_msg
            )

    def create(self, name, working_dir):
        """Create a new project object.

        Args:
            name (str): project name
            working_dir (str): full path to the project
        """
        self._project = {
            "name": name,
            "acquisition": deque(),
            "working_directory": working_dir,
        }

    @staticmethod
    def change_name(obj, name):
        """Change the name of an Acquisition or an Event.

        Args:
            obj (Event or Acquisition): object to change name on.
            name (str): Name as a string.
        """
        if isinstance(obj, type(Acquisition())):
            obj.name = name
        elif isinstance(obj, dict):
            obj["name"] = name

    def add_acquisition(self, name=None):
        """Add a new aquisition to the project.

        Acquisition is added to the top of the stack.

        Args:
            name (str): Default is "Acquisition n", n is the number of prev. Acquisitions.

        """
        if not name:
            name = f"Acquisition {len(self._project['acquisitions'])}"
        acq = Acquisition(
            name=name,
            acq_num=len(self._project["acquisitions"]),
            maxlen=self.maxlen,
        )
        if self._project.get("acquisitions", None) is None:
            self._project["acquisitions"] = deque()
        self._project["acquisitions"].append(acq)

    def get_acquisition(self, value: int = -1) -> dict:
        """Return the acquisition.

        Returns the last acquisition of no number is specified.

        Args:
            value (str): Acquisition number, default -1 (last)
        Returns:
            Acquisition
        """
        if not isinstance(value, int):
            raise TypeError("Value need to be int")

        if not self._project.get("acquisitions", None):
            LOGGER.debug("Found no acquistition using get, adding acquisition.")
            self.add_acquisition()

        if value > len(self._project["acquisitions"]):
            value = len(self._project["acquisitions"]) - 1

        if value < -len(self._project["acquisitions"]):
            value = -len(self._project["acquisitions"]) + 1

        return self._project["acquisitions"][value]

    def get_active_acquisition(self):
        """Returns the active acquisition.

        Active acquisition is used for plotting, printing and updating of acquisitions.
        For adding new events, use get_acquisition, it returns the last one.
        If no active acquisition has been set, returns the last one.
        """

        if not self._active_acquisition:
            self._active_acquisition = self.get_acquisition()
        return self._active_acquisition

    def set_active_acquisition(self, acquisition=None):
        """Set the active acquisition.

        Doesn't have to be from the acquisitions deque, can be any acquisition object.
        active_acquisition can be used independent of get_acquistion which is used to add events.
        If no acquisition is supplied, the last acquisition is set.
        """
        if not acquisition:
            self._active_acquisition = self.get_acquisition()
        else:
            self._active_acquisition = acquisition

    def amount_acquisitions(self):
        """The amount of aquisitions currently in the project.

        Returns:
            Amount of aquisitions as an integer.
        """
        return len(self._project.get("acquisitions"))

    def save(self, filename):
        """PICKLE THE ENTIRE PROJECT!"""
        if filename is None:
            raise TypeError("Supplied pathname is NoneType.")
        path, _ = os.path.split(filename)
        if not os.path.isdir(path):
            raise NotADirectoryError(f"Not a valid directory: {path}")

        try:
            sfile = gzip.GzipFile(filename, "w", compresslevel=4)
            pickle.dump(self._project, sfile, protocol=pickle.HIGHEST_PROTOCOL)
        except IOError as error_msg:
            raise ProjectIOError("File could not be created. %s", error_msg)
        except pickle.PicklingError as error_msg:
            raise ProjectIOError("Saving acquisition failed: %s", error_msg)

    def load(self, filename):
        """Loads a saved Project.

        Args:
            filename (str): Full path to the file including the name of the file.
        """
        if not os.path.isfile(filename):
            raise FileNotFoundError(f"No file found: {filename}")
        try:
            sfile = gzip.GzipFile(filename, "r")
            project = pickle.load(sfile)
        except IOError as error_msg:
            raise ProjectIOError("File could not be created. %s", error_msg)
        except pickle.UnpicklingError as error_msg:
            raise ProjectIOError("Loading acquisition failed: %s", error_msg)
        else:
            if not isinstance(project, type(dict())):
                raise TypeError("Not a valid project file.")
            self._project = project

    @staticmethod
    def save_acq(acq, filename):
        """Save an Acquisition.
        Takes an Acquistion and a filename.
        Will overwrite existing file without asking.

        Args:
            acq (Acquisition): Acquisition to save
            filename (path): Full filepath and filename to save.
        """
        if not isinstance(acq, type(Acquisition())):
            raise TypeError("Supplied Acquisition is NoneType.")
        if filename is None:
            raise TypeError("Supplied pathname is NoneType.")

        path, _ = os.path.split(filename)
        if not os.path.isdir(path):
            raise NotADirectoryError(f"Not a valid directory: {path}")

        try:
            save_pickle_acquisition(acq, filename)
        except IOError as error_msg:
            raise AcquisitionIOError("File could not be created. %s", error_msg)
        except pickle.PicklingError as error_msg:
            raise AcquisitionIOError("Saving acquisition failed: %s", error_msg)

    def load_acq(self, filename):
        """Load an acquisition to the current project.

        Args:
            filename (path): Valid acquisition file.

        Raises:
            NotADirectoryError if a non-valid path is supplied.
            PicklingError if the pickle failed.
            IOError if the file can't be read.
        """

        path, _ = os.path.split(filename)
        if not os.path.isdir(path):
            raise NotADirectoryError(f"Not a valid directory: {path}")

        try:
            acq = load_pickle_acquisition(filename)
        except IOError as error_msg:
            raise AcquisitionIOError("File could not be loaded. %s", error_msg)
        except pickle.UnpicklingError as error_msg:
            raise AcquisitionIOError("Loading acquisition failed: %s", error_msg)
        else:
            self._project["acquisitions"].append(acq)

    def delete(self, item):
        """Delete acquisition or event from project.

        Args:
            item (tuple): (acq_num, evt_num)
        """
        if item[1] is None:
            self.delete_acq(item)
        else:
            self.delete_evt(item)

    def delete_evt(self, item: tuple):
        """Remove event from Acquisition.

        Args:
            item (tuple): tuple with (Acquisition #, Event #).
        """
        acquisition = self._project["acquisitions"][item[0]]
        event = item[1]

        acquisition.rotate(-event)
        acquisition.popleft()
        acquisition.rotate(event)

    def delete_acq(self, item: tuple):
        """Delete the acquisitions in the project.

        Args:
            item (tuple): tuple with (Event, Acquisition)
        """
        index = item[0]
        self._project["acquisitions"].rotate(-index)
        self._project["acquisitions"].popleft()
        self._project["acquisitions"].rotate(index)

    @staticmethod
    def _find_event_in_acquisition(event, acquisition):
        """Find an event for deletion using the index rather than the built-in.

        deques index and remove function can't find numpy objects properly.
        Give more flexibility by finding using object id rather than the built in.

        Args:
            event: item to find
            acquistion: deque-like object to find item in.

        Returns:
            index int if found or None if not found.
        """
        for index, item in enumerate(acquisition):
            if id(event) == id(item):
                return index

        return None

    def __getitem__(self, value):
        if value is None:
            raise IndexError

        if isinstance(value, int):
            return self._project["acquisitions"][value]

        if len(value) > 2:
            raise IndexError

        if isinstance(value[0], slice) or isinstance(value[1], slice):
            TypeError("Project doesn't support slices yet")

        acq_idx = value[0]
        evt_idx = value[1]

        if evt_idx is None:
            return self._project["acquisitions"][acq_idx]

        return self._project["acquisitions"][acq_idx][evt_idx]

    def __len__(self):
        return len(self._project["acquisitions"])

    def __bool__(self):
        return len(self) != 0

    def __iter__(self):
        yield from self._project["acquisitions"]



"""Timing correction calibration

Tool to generate the timing data.

To generate the timing data, data should be captured with a setup as shown in
link:

DESCRIPTION

A large number of events need to be captured, a minimum of 10k events is required.
The data must be captured using

"""
import logging
from copy import deepcopy
from typing import List

import numpy as np

from naludaq.helpers.exceptions import BadDataError

LOGGER = logging.getLogger(__name__)


class TimingCalibration:
    def __init__(self, board):
        """Create calibration data for timing calibration."""
        self.board = board
        self.progress = []
        self.num_xings = 2  # Assume two peaks
        self.threshold = 300
        self.xing_num = None
        self.samp_corr = False
        self._minimum_events = 50  # _000
        self._cancel = False

    @property
    def progress(self):
        """Get/Set the progress message queue.

        This is a hook the read the progress if running threads.
        """
        return self._progress

    @progress.setter
    def progress(self, value):
        if not hasattr(value, "append"):
            raise TypeError(
                "Progress updates are stored in an object with an 'append' method"
            )
        self._progress = value

    @property
    def minimum_events(self):
        """Get/Set the minimum events needed to run calibration

        This is used to validate a dataset.
        """
        return self._minimum_events

    @minimum_events.setter
    def minimum_events(self, value):
        if not isinstance(value, int):
            raise TypeError("minimum_events must be an integer")
        if value <= 0:
            raise ValueError("minimum_events must be a positive value")
        self._minimum_events = value

    def cancel(self):
        self._cancel = True

    def generate(self, events, channel: int = 0, threshold: int = 300, peaks=(0, 1)):
        """Generate timing calibration and adds to board.

        Will validate the incoming data and raise and error if
        it doesn't contain enough data to run a validation

        Args:
            events: collection of events
            channel: channel containing data
            threshold: int value to intersect peaks.
        """
        LOGGER.info(
            "Generate timing cal constants using, %s events, ch: %s, threshold: %s",
            len(events),
            channel,
            threshold,
        )
        old_timingcal = deepcopy(self.board.timingcal)

        events = self._get_valid_events(events, channel, threshold, peaks)
        if len(events) < self.minimum_events:
            self.board.timingcal = old_timingcal
            raise BadDataError("Not enough valid events to run calibration")

        dt = self._try_generate_dts_or_log(events, threshold, channel)

        if self._cancel is True:
            return

        self.board.timingcal = dt
        self.progress.append(("Timing calibration set successfully", 100))
        LOGGER.debug("Timing calibration set successfully")
        return

    def _try_generate_dts_or_log(self, events, threshold, channel):
        dt = None
        try:
            dt = self.generate_dts(events, threshold)[channel]
        except Exception as emsg:
            import traceback

            track = traceback.format_exc()
            LOGGER.debug("Failed to generate dts due to %s\n%s", emsg, track)
            self._cancel = True
        return dt

    # VALIDATION ###################################################
    def _get_valid_events(
        self, events: dict, channel: int, threshold: int, peaks: tuple
    ) -> List[dict]:
        """Measure distance between two pulses using simple threshold

        Will skip the event if more or less than 2 peaks.

        Args:
            events: list of events to validate
            channel: channnel of interest
            threshold: integer for intersection with peak
            corrected: timing correct time field in event.
        Returns:
            list of distances between peaks.
        """
        indexed_distances = self._distances_between_two_peaks(
            events, channel, threshold, peaks
        )
        if indexed_distances:
            output_idx = self._drop_distance_outliers(indexed_distances)
        else:
            output_idx = []

        output = [events[i] for i in output_idx]

        return output

    def _find_all_rising_edge_xings(self, event: dict, channel: int, threshold: int):
        """Returns all rising edge xings in an event.

        Args:
            event: collection of events to check
            threshold: threshold value to use for the check.
            num_xings: number of crossing to expect

        Returns:
            List of all rising crossings between event data and threshold
        """
        return np.where(np.diff(np.sign(event["data"][channel] - threshold)) > 0)[0]

    def _distances_between_two_peaks(
        self, events: dict, channel: int, threshold, peaks: tuple
    ):  # -> List[int, float]:
        """Return a list[(index, distance)] for all distances."""
        distances = []
        for idx, evt in enumerate(events):
            if "data" not in evt.keys():
                continue
            xings = self._find_all_rising_edge_xings(evt, channel, threshold)
            if len(xings) == 0:
                continue
            xaxis = evt["time"][channel]
            x_0 = xaxis[xings[peaks[0]]]
            x_1 = xaxis[xings[peaks[1]]]
            dx = x_1 - x_0

            distances.append((idx, dx))
        return distances

    def _drop_distance_outliers(self, distances, deviation: float = 5.0):
        """Returns all distances within the tolerated interval

        Args:
            distances (list): List of tuples with (index, distance)
            deviation: distance in sample counts from the mean distance

        Returns:
            List of indexes that pass the test
        """

        mean = np.mean(np.array(distances)[:, 1])
        output = []
        for d in distances:
            if (mean - deviation) < d[1] < (mean + deviation):
                output.append(d[0])
        return output

    ################################################################

    def generate_dts(self, events, threshold=600):
        """Generates a `dt` array which is used to correct events.

        Combines combile_bof() and calc_dts() into a single function
        Determine the locations of the crossings

        Args:
            events (deque): List of events, see module description for setup.
            threshold (int): threshold value used.
        """
        samples = self.board.params.get("samples", 64)
        bof = self._compile_bof(events, threshold=threshold)
        results = list()

        for i, chan_bof in enumerate(bof):
            if np.all(np.isnan(chan_bof)):
                results.append(np.zeros(samples * 2))  # TODO: Assumes 128 samples.
            else:
                dts = self._calc_dts(chan_bof)
                results.append(dts)

        return results

    def _compile_bof(self, events, **kwargs):
        """
        Inputs a bunch of events with a randomly occuring pulse and determines
        the locations of the crossings

        Returns a list *for each channel*
        """
        channels = self.board.params.get("channels", 4)
        results = list()
        for chan in range(channels):
            results.append(list())
        for event in events:
            try:
                xings = self._find_bin_edge(event, **kwargs, samp_corr=True)
            except Exception as err:
                LOGGER.error("Couldn't find bin edge:%s", err)
                continue
            for chan, chan_xings in enumerate(xings):
                for xing in chan_xings:
                    results[chan].append(xing)
            if self._cancel is True:
                return []

        aresults = list()
        for result in results:
            # print(f"LEN RESULT: {len(result)}")
            aresults.append(np.array(result))

        return aresults

    def _find_bin_edge(self, event, **kwargs):
        if "threshold" in kwargs.keys():
            threshold = kwargs["threshold"]
        else:
            threshold = self.threshold  # 600
        if "xing_num" in kwargs.keys():
            xing_num = kwargs["xing_num"]
        else:
            xing_num = self.xing_num  # None

        if "samp_corr" in kwargs.keys():
            samp_corr = kwargs["samp_corr"]
        else:
            samp_corr = self.samp_corr  # False

        all_chan_xings = list()

        # loop over channels
        for data, winds in zip(event["data"], event["window_labels"]):
            chan_xings = list()

            # possible there isn't any data i guess
            if data is None:
                chan_xings.append(np.nan)
            else:
                xings = np.where(np.diff(np.sign(data - threshold)) > 0)[0]
                # print(f"{xings=}")
                if len(xings) <= 1:  # Can't use one peak
                    chan_xings.append(np.nan)
                    continue
                if (
                    len(xings) != self.num_xings
                ):  # Default two peaks, but no reason to run more
                    xings = xings[:2]
                    # print(f"{xings=}")
                    #  chan_xings.append(np.nan)
                # else:
                if xing_num is None:
                    for xing in xings:
                        # print(f"MEEP {xing=}")
                        if samp_corr:
                            value = self._xing_to_time_pos(xing, winds[0])
                        else:
                            value = xing
                        chan_xings.append(value)
                else:
                    if samp_corr:
                        value = self._xing_to_time_pos(xings[xing_num], winds[0])
                    else:
                        value = xings[xing_num]
                    chan_xings.append(value)
                # print(f"{chan_xings=}")
            all_chan_xings.append(chan_xings)

        return all_chan_xings

    def _xing_to_time_pos(self, xing, first_wind):
        samples = self.board.params.get("samples", 64)
        return (xing + (first_wind % 2) * samples) % (samples * 2)

    def _calc_dts(self, crossings):
        """
        Inputs a list containing the locations of zero crossings

        Calculates the timing calibration from this assuming that you expect an even distrubution

        Only for a *single channel*
        """
        bins = 2 * self.board.params.get("samples", 64)
        dt = (
            np.histogram(crossings % bins, bins, (-0.5, bins - 0.5), density=True)[0]
            * bins
        )
        positions = np.insert(np.cumsum(dt), 0, 0)[:-1]

        return positions - np.arange(0, bins)



import functools
import logging


def log_function(logger=None, level=logging.INFO):
    """Decorator that logs a noticable visual separator before
    running the decorated function. Useful for getting log output
    organized.

    Args:
        logger (Logger): logger to use when displaying the separator.
            If no logger is supplied, the text is printed instead.
        level (int): logging level to use if a logger is supplied.
    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            message = f" Running: {func.__name__} ".ljust(80, "#")
            if logger is not None:
                logger.log(level, message)
            else:
                print(message)
            return func(*args, **kwargs)

        return wrapper

    return decorator



"""Facade for all the underlying controllers.

This is the class to import to talk to all aspects of the board.
It contains all the aspects of the board.
Talk to the board,
receive data
setup connection
generate pedestals
storing data.
"""
import os
import time
from collections import deque
from copy import deepcopy
from logging import getLogger

from naludaq.backend import AcquisitionManager, RemoteAcquisition
from naludaq.backend.exceptions import BackendError
from naludaq.board import init_hardware, startup_board
from naludaq.controllers import get_board_controller, get_readout_controller
from naludaq.daq import LightDaq as DataAcquisitionController
from naludaq.daq.hiperdaq import HiperDaq as hDAQ
from naludaq.daq.preprocess import Preprocess
from naludaq.helpers import event_transfer_time
from naludaq.io import io_manager
from naludaq.tools.pedestals import get_pedestals_controller
from naludaq.tools.waiter import EventWaiter

LOGGER = getLogger(__name__)


class NaluDaq:
    """High-Level interface for the Naludaq using interactive or GUI.

    Creates an interface to the different components required to run the board.

    Args:
        board_model(str): The model of the board.
        working_dir(str): location on disk to store data.
    """

    def __init__(self, board, working_dir, start_workers=True):
        self.board = board
        if not os.path.isdir(working_dir):
            raise NotADirectoryError(
                f"NaluDaq init, directory not found: {working_dir}"
            )

        if not self.board.using_new_backend:
            self.daq = DataAcquisitionController(board, working_dir, start_workers)
            self.daq.preprocess = Preprocess(board)

        self.readout_settings = {
            "trig": "ext",
            "lb": "trigrel",
            "acq": "raw",
            "ped": "zero",
            "readoutEn": True,
            "singleEv": False,
        }

        self.is_capturing = False
        self.is_connected = False

    def init_board(self) -> bool:
        """Initializes hardware.

        Runs through the board initialization sequence.
        This initialization is board/model specific and requires the correct boardparams.
        It also requires an active connection with the board.

        Returns:
            True if the init is successful.
        """
        return init_hardware(self.board)

    def reset_board(self):
        """Reset the board.

        Sends a digital reset command to the board and resets the fpga.
        It will not wipe any settings made by the user.
        """
        get_board_controller(self.board).digital_reset()
        time.sleep(0.25)
        get_board_controller(self.board).digital_reset()

    def generate_pedestals(self):
        """Generate pedestals.

        Generate pedestals for the supplied board.
        """
        get_pedestals_controller(self.board).generate_pedestals()

    def save_current_pedestals_to_disk(self, filename):
        """Save current generated pedestals.

        Args:
            filename(str): full, correct path.
        """
        self.save_pedestals_to_disk(self.board.pedestals, filename)

    def save_pedestals_to_disk(self, pedestals, filename):
        """Save pedestals.

        Args:
            filename(str): full, correct path.
        """
        get_pedestals_controller(self.board).save_pedestals(pedestals, filename)

    def load_pedestals_from_disk(self, filename):
        """Save pedestals.

        Args:
            filename(str): full, correct path.
        """
        return get_pedestals_controller(self.board).load_pedestals(filename)

    def export_pedestals_as_csv(self, pedestals, filename):
        """Save pedestals.

        Args:
            pedestals(Pedestals): Pedestals object to save.
            filename(str): full, correct path.
            time_per_channel (bool): export time per channel rather than per row.
        """
        io_manager.IOManager().export_pedestals_csv(self.board, filename, pedestals)

    def reset_pedestals(self):
        """Reset pedestals data, the pedestals and set the pedestals to None."""
        self.board.pedestals = None

    def change_trigger_value(self, channel, value):
        """Set the trigger ADC for a specific channel.

        Trigger value is the value the board will trigger on
        in triggered mode.

        By setting this value !=0 the board will trigger on that channel.
        It's possible to trigger on multiple channels.

        Args:
            channel(int): Channel number to trigger on.
            value(int): Value to trigger on.
        """
        self.board.trigger.values[channel] = value

    def update_triggers(self):
        """Send all the trigger settings to the board."""
        self.board.trigger.write_triggers()

    def set_readout_settings(self, readout_settings):
        """Updates all the trigger values at once

        Args:
            readout_settings(dict):
                trig:
                lb:
                acq:
                ped:
                readoutEn:
                singleEv:
        """
        self.readout_settings = readout_settings

    def start_acquisition(self, readouts=0):
        """Start acquiring data from the board and store it.

        Make sure to swap the output storage before starting.

        Args:
            readouts(int): Maximum amount of readouts, 0 = infinite
        """
        self.update_triggers()
        if readouts > 0:
            self.readout_settings["singleEv"] = True
            get_readout_controller(self.board).number_events_to_read(readouts)
        else:
            self.readout_settings["singleEv"] = False

        self.is_capturing = True
        if not self.board.using_new_backend:
            self.reset_pkg_num()
            self.daq.start_capture()

        get_board_controller(self.board).start_readout(**self.readout_settings)

    def stop_acquisition(self):
        """Stop acquiring data."""
        try:
            get_board_controller(self.board).stop_readout()
        except Exception as error_msg:
            LOGGER.warning("stop_acquisition can't stop due to: %s", error_msg)
        self._stop_capture()

    def _stop_capture(self):
        if self.board.using_new_backend:
            try:
                AcquisitionManager(self.board).current_acquisition = None
            except BackendError as e:
                LOGGER.error("Failed to stop capture: %s", e)
        else:
            self.daq.stop_capture()
            LOGGER.debug("Buffers: %r", self.daq.get_buffer_levels())
        self.is_capturing = False

    def change_output(self, acquisition):
        """Set the output storage.
        Appends parsed events to the storage.
        Can be either a deque or an Acquisition.

        """
        if self.board.using_new_backend:
            # TODO
            pass
        else:
            pedestals = deepcopy(self.board.pedestals)
            caldata = deepcopy(self.board.caldata)
            acquisition.caldata = caldata
            acquisition.pedestals = pedestals
            self.daq.switch_output_buffer(acquisition)

    def peek_latest_event(self) -> dict:
        """Attempt to retrieve the latest event received.

        Raises:
            IndexError: if the last event could not be read.

        Returns:
            dict: the raw event
        """
        event = None
        if self.board.using_new_backend:
            acq = AcquisitionManager(self.board).current_acquisition
            if acq is None:
                raise IndexError("Current Acquisition is not set.")
            try:
                event = acq[-1]
            except IndexError as e:
                raise e
        else:
            try:
                event = self.daq.peek_latest_event()
            except IndexError as e:
                raise e
        return event

    def change_disk_storage_directory(self, directory):
        """Save data to this folder.

        Args:
            directory: Path to a directory on disk.

        Raises:
            NotADirectoryError if directory is not found.
        """
        if self.board.using_new_backend:
            return
        if not os.path.isdir(directory):
            raise NotADirectoryError(f"{directory} is not a valid directory.")

        self.daq.storage_directory = directory
        LOGGER.debug("Storage directory set to: %s", directory)

    def enable_disk_storage(self, enable):
        """Enable/disable store raw/parsed to disk.

        Activate the processes to store raw packages and/or parsed data to disk.
        The default directory is the project directory.
        Starts a separate thread for each data type, tapping in on the data pipeline
        and storing to the disk.

        Args:
            to_store(str): what to enable/disable "raw" or "parsed" storage
            enable(bool): Set True to enable.
        """
        if self.board.using_new_backend:
            return
        self.daq.store_to_disk = enable
        LOGGER.debug("Naludaq enable disk storage: %s", enable)

    def reset_pkg_num(self):
        """Reset the package number counter"""
        if not self.board.using_new_backend:
            self.daq.pkg_num = 0

    def update_storage_settings(self, storage_settings):
        """Updates the flags used when storing events to disk, these flags
        sets what the Preprocess does to raw events. The dict can contain
        any combination of keys.

        Args:
            storage_settings (dict): {
                correct_pedestals:
                convert_adc2mv:
                correct_timing:
                convert_samples2time:
                process_in_place:
            }
        """
        if self.board.using_new_backend:
            return
        self.daq.storage_settings = storage_settings

    def select_readout_channels(self, channels_to_read):
        """Set the channels to read on the board

        Args:
            channels_to_read: List of channel numbers to readout.

        """
        get_readout_controller(self.board).set_readout_channels(channels_to_read)

    def change_board(self, board):
        """Resetting connection to hardware, reset data acquisition and change connection.

        It will reset the connection even if the new connection params are
        the same as the old parameters.
        Need to reconnect after a change of the board.

        Args:
            board_model(str): Boardmodel, if none is given, demoboard is assumed.
        """
        if self.is_connected:
            self.disconnect()
            if self.board.using_new_backend:
                AcquisitionManager(self.board).current_acquisition = None
            else:
                self.daq.stop_capture()
            self.is_connected = False

        # update board_controller, reinitialize
        self.board = board

        # redistribute board controller
        try:
            self.daq.board = self.board
        except Exception:
            pass

    def connect(self, connection_params) -> bool:
        """Connect to the board using the set connection settings.

        Args:
            connection_params(dict):
                {"type": type of connection
                "model": Board model
                "usb_port": COM port or ttyUSB
                "usb_rate": Baud rate for uart
                "ip_addr": ip address XXX.XXX.XXX.XXX
                "ip_port": port}

        Returns:
            True if connection.
            False if connection failed.
        """
        # TODO(Marcus): Logic might be outdated, board carries to conection_params, check if this is a valid function.
        try:
            self.is_connected = startup_board(self.board)
        except (AttributeError, ConnectionError) as e_msg:
            LOGGER.error("Connect(): %s", e_msg)
            self.is_connected = False

        return self.is_connected

    def disconnect(self):
        """Disconnect gracefully from the board.

        If the connection is a serial connection it will try and lower the
        baudrate before closing.
        This will prevent the board to listen at 3M baud and the startup
        baudmatching will be much quicker.
        """
        try:
            self.board.disconnect()
        except AttributeError:
            pass
        self.board.connection = None
        self.is_connected = False
        LOGGER.debug("Board disconnected.")

    def quit(self):
        """Close connection and stop all threads/processes."""
        self.stop_acquisition()
        if not self.board.using_new_backend:
            self.daq.stop_workers()
        self.disconnect()


class UpacDaq(NaluDaq):
    """Special HACK adapter for the upac board.

    Need to be cleaned up with a proper factory.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.trigger_mode = "00"
        self.read_amount = 0

    def capture_event(self):
        """ARGH, UGLY HACK Function for software trigger!

        Seriously, AARGHH!

        """
        self.is_capturing = True
        bc = get_board_controller(self.board)
        bc.clear_buffer()
        new_buffer_size = self._current_output_buffer_len() + 1
        if not self.board.using_new_backend:
            self.daq.start_capture()

        time.sleep(0.1)
        bc.start_readout(self.trigger_mode)
        bc.toggle_trigger()
        bc.stop_readout()
        timeout = event_transfer_time(
            self.board,
            self.board.params["windows"],
            overhead=3,
        )
        waiter = EventWaiter(
            buffer=self._output_buffer(),
            amount=new_buffer_size,
            interval=0.2,
            timeout=timeout,
        )
        LOGGER.debug("Using timeout for capture: %s sec", timeout)

        try:
            waiter.start(blocking=True)
        except TimeoutError:
            raise
        finally:
            self._stop_capture()
            # self.read_amount = self.daq.pkg_num

    def reread_event(self):
        """"""
        new_buffer_size = self._current_output_buffer_len() + 1
        prev_capturing = self.is_capturing
        if not self.is_capturing:
            start_readout = self.board.model not in ["upac96"]
            self.start_acquisition(start_readout)
        get_board_controller(self.board).toggle_reread()  # toggle_trigger()

        timeout = event_transfer_time(
            self.board, self.board.params["windows"], overhead=3
        )
        waiter = EventWaiter(
            buffer=self._output_buffer(),
            amount=new_buffer_size,
            interval=0.2,
            timeout=timeout,
        )
        LOGGER.debug("Using timeout for capture: %s sec", timeout)

        try:
            waiter.start(blocking=True)
        except TimeoutError:
            raise
        finally:
            if not prev_capturing:
                self.stop_acquisition()
        # self.read_amount = self.daq.pkg_num
        # self.is_capturing = False

    def start_acquisition(self, start_readout: bool = True):
        """Start acquiring data from the board and store it.

        On certain boards starting the readout blocks the re-read functionality.
        Consult the firmware/ASIC documentation! (What documentation..?)
        Make sure to swap the output storage before starting.

        Args:
            start_readout (bool): whether to send the command to start
                the readout.
        """
        LOGGER.debug("STARTING, %s", self.trigger_mode)

        self.is_capturing = True
        self.reset_pkg_num()
        if not self.board.using_new_backend:
            self.daq.start_capture()
        if start_readout:
            get_board_controller(self.board).start_readout(self.trigger_mode)

    def stop_acquisition(self):
        """Stop acquiring data."""
        super().stop_acquisition()
        self.reset_pkg_num()

    def _output_buffer(self) -> "deque | RemoteAcquisition":
        """Returns the output buffer

        Returns:
            "deque | RemoteAcquisition": the output buffer
        """
        if self.board.using_new_backend:
            return AcquisitionManager(self.board).current_acquisition
        else:
            return self.daq.output_buffer

    def _current_output_buffer_len(self) -> int:
        """Get the current number of events in the output buffer"""
        if self.board.using_new_backend:
            return len(AcquisitionManager(self.board).current_acquisition or [])
        else:
            return len(self.daq.output_buffer)


class HiperDaq(NaluDaq):
    """Special HACK adapter for the upac board.

    Need to be cleaned up with a proper factory.
    """

    def __init__(self, board, working_dir, *args, start_workers=True, **kwargs):
        super().__init__(board, working_dir, start_workers=True)
        if not self.board.using_new_backend:
            self.daq = hDAQ(self.board, working_dir, start_workers)
            self.daq.preprocess = Preprocess(board)
        self.chips: list = [0]
        self.read_amount = 0
        self.readout_settings = {
            "trig": "ext",
            "lb": "forced",
            "acq": "raw",
            "dig_head": False,
            "ped": "zero",
            "readoutEn": True,
            "singleEv": False,
        }

    def select_readout_chips(self, chips_to_read):
        self.chips = chips_to_read

    def reset_board(self):
        """Reset the board.

        Sends a digital reset command to the board and resets the fpga.
        It will not wipe any settings made by the user.
        """
        get_board_controller(self.board).reset_board()

    def capture_event(self):
        """Send a trigger pulse to the board"""
        get_board_controller(self.board).toggle_trigger()

    def start_acquisition(self):
        """Start acquiring data from the board and store it.

        Make sure to swap the output storage before starting.

        Args:
            readouts(int): Maximum amount of readouts, 0 = infinite
        """
        self.is_capturing = True
        bc = get_board_controller(self.board)
        self.daq.pkg_num = 0
        self.daq.start_capture()
        bc.start_readout(**self.readout_settings)

    def stop_acquisition(self):
        """Stop acquiring data."""
        if self.board.connection is not None:
            try:
                get_board_controller(self.board).stop_readout()  # Nuke the board
            except Exception as error_msg:
                LOGGER.warning("stop_acquisition can't stop due to: %s", error_msg)
        self.daq.stop_capture()
        self.daq.pkg_num = 0
        self.is_capturing = False
        self.daq.get_buffer_levels()



"""
Driver for talking over a UART interface

Should be interoperable with the linketh package


Ben Rotter - Nalu Scientific
email: ben@naluscientific.com
September 2018
"""
import logging
import time

import numpy as np
import serial
import serial.tools.list_ports

from naludaq.board.connections.base_serial import BaseSerialConnection
from naludaq.helpers import validations
from naludaq.parsers.answer_parser import get_answer_parser

logger = logging.getLogger("naludaq.UART")


class UART(BaseSerialConnection):
    """UART is a communications class for talking to the boards."""

    def __init__(self, conn_info):
        super().__init__()
        self._write_addr = 0xAF
        self._read_addr = 0xAD
        self._baud = conn_info["speed"]
        self._model = conn_info["model"]
        self._rxtx_timeout = 0.1  # s

        self.stopword = conn_info.get("stop_word", b"\xca\xfe")  # b'\xfa\xce'
        self.new_firmware = conn_info.get("new_firmware", False)
        self.response_length = 4
        self.parse_answer = get_answer_parser()

        serial_number = conn_info.get("serial_number", None)
        com_port = conn_info.get("usb_addr", None)

        if serial_number is not None:
            logger.debug("using serial_number %s to connect", serial_number)
            self.serial_number = serial_number
        elif com_port is not None:
            logger.debug("using usb_addr %s to connect", com_port)
            self.port = com_port
        else:
            raise ConnectionError(
                "Need either serial number or COM port to open connection"
            )

        self.stopword = conn_info["stop_word"]  # b'\xfa\xce'
        self.ser = None

    def __del__(self):
        self.close()

    @property
    def baud(self):
        """Get/Set baudrate.

        Must be a positive integer.
        Raises:
            ValueError
        """
        return self._baud

    @baud.setter
    def baud(self, value):
        if isinstance(value, int) and value >= 0:
            self._baud = value
        else:
            raise ValueError("Baudrate must be a positive integer.")
        if self.ser:
            try:
                self.ser.baudrate = value
            except Exception as e:
                logger.error("Failed to set baud rate: {e}")
                raise ConnectionError("Failed to set baud rate due to error: {e}")

    @property
    def port(self):
        return self._port

    @port.setter
    def port(self, value):
        if not isinstance(value, str):
            raise TypeError("Port name must be a string.")
        ports = self.get_available_ports()
        logger.debug("available ports:%s", ports)

        if value not in ports:
            raise ValueError(f"{value} is not a valid port.")
        self._port = value

    @property
    def serial_number(self):
        return self._serial_number

    @serial_number.setter
    def serial_number(self, value):
        """
        Set a port given a part of a serial number
        """

        if not isinstance(value, str):
            raise TypeError("Serial Number must be a string.")

        ports = serial.tools.list_ports.comports()
        logger.debug("Connected Devices:")
        for port in ports:
            logger.debug("sn:%s, port:%s", port.serial_number, port.device)
            if port.serial_number is None:
                continue
            if value in port.serial_number:
                self._serial_number = port.serial_number
                self._port = port.device
                logger.debug("found port matching given SN with %s", value)
                return

        logger.error("did not find port matching SN")

    def get_available_ports(self):
        """Lists available serial ports.

        Returns:
            A list of available serial ports on the system. Empty if no ports are available.
        """
        return [comport.device for comport in serial.tools.list_ports.comports()]
    
    @property
    def com_port(self):
        """Get/set the COM port.
        Alias of UART.port to use the same naming convention as the backend.

        Raises:
            TypeError if set to a non-integer type.
            ValueError if set to a device that is not available.
            ConnectionError if set while the connection is open
        """
        return int(self.port.lstrip("COM"))
    
    @com_port.setter
    def com_port(self, value):
        if not isinstance(value, int):
            raise TypeError("COM port must be an integer")
        self.port = "COM{}".format(value)

    @property
    def type(self):
        return "uart"

    @property
    def is_open(self) -> bool:
        """Indicates whether the connection is open."""
        is_open = self.ser.is_open
        try:
            self.ser.in_waiting
        except:
            is_open = False
        return is_open

    def open(self):
        """Open connection over the UART interface.

        Set port and baud before opening a connection, use:
            .port(value)
            .baud(value)

        """
        if self.ser:
            if not self.ser.is_open:
                self.ser.open()

            return self.ser

        serial_connection = None
        try:
            serial_connection = serial.Serial(
                self.port, self._baud, timeout=self._rxtx_timeout
            )

        except ValueError as e_msg:
            logger.error("UART.open raise: %s", e_msg)
            serial_connection = None
        except serial.SerialException as error_msg:
            logger.debug(error_msg)
            serial_connection = None
        else:
            logger.info("Success!  Connected. %s", self.port)
        self.ser = serial_connection

    def close(self):
        """Close the serial port."""
        try:
            self.ser.close()
        except:
            pass

    def send(self, data):
        """Converts the string to a series of bytes to send in the correct format

        Input:
            data (hex): A hex type string (interface needs 32 bits, so 8 hex chars minimum)
            pause (float): How logn to wait in between send.
        """
        # logger.debug("Send(): data: %s", data)
        numchars = len(data)  # number of hex characters
        numcmds = numchars // 8  # 8 hex chars per command
        extra = numchars % 8

        if extra != 0:
            logger.debug(
                "UART.send(): need multiples of 8 hex chars, you gave me %s", numchars
            )
            logger.debug("command:%s", data)
            return

        logger.debug("UART.send(): %s", data)

        tosend = []
        for i in range(0, numchars // 2):
            tosend.append(int(data[i * 2] + data[i * 2 + 1], 16))

        try:
            self.ser.write(bytearray(tosend))
        except AttributeError as error_msg:
            logger.error("Can't send command to board, no connection available.")

    def toggleLoopback(self, state):
        """
        Turns the loopback on or off
        """
        command = "7009BAC" + str(int(state))
        self.send(command)
        try:
            self.ser.read(4)
        except:
            pass

    def read(self, num_bytes) -> bytes:
        return self.ser.read(num_bytes)

    def read_all(self) -> bytes:
        """Read all waiting bytes"""
        return self.ser.read_all()

    def reset_input_buffer(self):
        return self.ser.reset_input_buffer()

    def read_until(self, stopword):
        """Pulls data from the connection until a set of characters is found.

        Args:
            stopword (bytes): the sequence of characters that indicate
                when to stop reading.

        Returns:
            A bytearray containing the data, including the trailing stop word.
            The buffer is empty or incomplete if timed out
        """
        lenstop = len(stopword)
        buff = bytearray()
        while True:
            try:
                value = self.ser.read(1)
            except Exception as e:
                logger.debug("Receiving data failed due to: %s", e)
                break

            if value == b"":
                logger.debug("Receive timeout, no data returned")
                break

            buff += value
            if buff[-lenstop:] == stopword:
                break

        logger.debug("Received: %s", buff.hex())

        return buff

    @property
    def in_waiting(self):
        return self.ser.in_waiting

    def receive(self):
        """Reads out a package. Packages end with a stopword."""
        return self.read_until(self.stopword)

    @property
    def rtscts(self) -> bool:
        """Get/set whether RTS/CTS flow control is enabled."""
        return self.ser.rtscts

    @rtscts.setter
    def rtscts(self, value: bool):
        self.ser.rtscts = value

    def set_answer_parser(self, old=True):
        """Set's the answer parser"""
        self.parse_answer = get_answer_parser(old)

    def readReg(self, regNum, timeout=10000):
        if self._model in ["siread", "upac32", "upaci", "zdigitizer"]:
            return self.readReg_old(regNum=regNum, timeout=timeout)
        else:
            return self.readReg_new(regNum=regNum, timeout=timeout)

    def readReg_old(self, regNum, timeout=10000):

        command = f"{self._read_addr:02X}{regNum:02x}0000"
        self.send(command)
        # while self.ser.in_waiting < 4:
        #     timeout -= 1
        #     if timeout < 0:
        #         logger.error("No response, timed out")
        #         return -1
        buff = self.receive()
        logger.debug("readReg() Buff: %s, type: %s", buff, type(buff))
        # struct.unpack(): 'H' is for unsigned short,
        # > is because the stat fifo returns things big-endian for who knows why
        value = 0
        if len(buff) == 4:
            try:
                header = int.from_bytes(buff[0:2], byteorder="big", signed=False)
                read_reg = header & 255  # only 8 lsb
                value = int.from_bytes(buff[2:4], byteorder="big", signed=False)
            except Exception:
                logger.error("couldn't unpack output from register: %s", regNum)
                raise Exception(
                    "readReg couldn't unpack the serial response."
                )  # TODO Change type

            if read_reg != regNum:
                logger.error(
                    "readReg didn't read expected register, wanted %s and got %s",
                    regNum,
                    read_reg,
                )
                return -1

        try:
            logger.debug("UART.readReg(%s) success: %s", read_reg, value)
        except Exception:
            pass

        return value

    def readReg_new(self, regNum, timeout=10000):
        """
        Reads out the register specified by the integer regNum
        Args:
            regNum (int): The registry number to read
            timeout (int): Amount of retries.
        Returns:
            Hex value of the serialread of two bytes.
        """
        # response_length = self.response_length

        # Send command
        self._send_readreg_cmd(regNum)

        # Wait for response
        buffer = self.receive()  # self._wait_for_readreg_response(response_length)
        # Parse response
        answer = self.parse_answer(buffer)
        # Make sure responses are valid

        try:
            logger.debug(
                "UART.readReg(%s) success: %s %s %s",
                answer["read_reg"],
                answer["value"],
                hex(answer["value"]),
                "{0:b}".format(answer["value"]).zfill(16),
            )
        except Exception:
            pass

        return answer["value"]

    def _send_readreg_cmd(self, regNum):
        cmd_id = np.random.randint(0, 2**16)
        if isinstance(regNum, str):
            regNum = int(regNum, 16)

        try:
            command = f"{self._read_addr:02x}{regNum:02x}{cmd_id:04x}"
        except Exception as error_msg:
            logger.exception(f"send_readreg_cmd create cmd failed: {error_msg}")
        else:
            self.send(command)

    def _wait_for_readreg_response(self, response_length, timeout: int = 300):
        """Polls the input buffer until there is a valid repsonse.

        Args:
            response_length: number of bytes in response (eg. 4 or 8)
            timeout (int): number of miliseconds before timing out.

        Returns:
            bytes returned.
        """
        buff = b""
        while self.ser.in_waiting < response_length:
            timeout -= 1
            time.sleep(0.001)
            if timeout < 0:
                logger.error("No readReg response, timed out")
                break

        # Serial reader can cause alignment issues, read all will fix (although first read fails)
        buff = self.ser.read_all()

        # Buffer can be too large if data sent after timeout code above.
        buff = buff[-response_length:]

        return buff

    def writeReg(self, regNum, value):
        """Writes a value to a register (16bits max)

        Args:
            regNum (int): registry number
            value: value to write
        """

        command = self.write_addr + hex(regNum)[2:].zfill(2) + hex(value)[2:].zfill(4)
        logger.debug("WriteReg: %s", command)
        self.send(command)

    def resync(self, sync_int, sync_addr) -> bool:
        """Resync the board if the input buffer is half full.

        Sometimes you go out of sync, and you need to send single chars till it works
        If the baud speed is set different you can end up with the boards input buffer being partially filled.
        Thus it waits for more characters, try send characters until it accepts an 8 char command again.

        Args:
            sync_int(int): value expected when synced.

        Returns:
            True if board is synced, False if syncing failed.
        """
        try:
            sync_regnum = int(sync_addr)
        except TypeError:
            sync_regnum = int(sync_addr, 16)
        self.ser.read_all()
        logger.debug("Syncing board...")
        for i in range(0, 8):
            value = b""
            try:
                value = self.readReg(sync_regnum)
                if value == b"":
                    continue
            except:
                logger.debug("UART:resync(): sending single character")
                self.ser.write(bytearray(b"F"))

            # val = binascii.hexlify(value).lstrip(b'0x').lower()
            if value == sync_int:
                logger.debug("synced, was off %s chars", i)
                return True
            else:
                continue

        logger.debug("Couldn't sync")
        return False

    def reset_input_buffer(self, iterations: int = 100):
        """Discards any data currently in the boards output buffer.

        This will empty both the computer input buffer and the hardware
        output buffer.

        However, this will only run 100 times at the most which is ~400kb.

        Args:
            iterations (int): number of times to empty the buffer.
                A cap is necessary since the board may be sending data faster
                than it can be purged.
        """
        validations.validate_positive_int_or_raise(iterations)
        # Buffer can refill with junk after being emptied,
        # need to empty it multiple times
        for _ in range(iterations):
            if self.in_waiting == 0:
                break
            self.ser.reset_input_buffer()

            # Without waiting, buffer refills right after the loop breaks
            time.sleep(0.01)

    def reset_output_buffer(self):
        """Discards any data currently in the output buffer.

        Discarded data will not be received by the board.
        """
        self.ser.reset_output_buffer()



import logging

from .peripherals_controller import PeripheralsController

logger = logging.getLogger("naludaq.peripherals_aardvarcv3")


class Aardvarcv3PeripheralsController(PeripheralsController):
    def __init__(self, board):
        """Aardvarcv3 peripherals controller.

        Provides a way to read various I2C devices on the board.

        Args:
            board (Board): the board object.
        """
        super().__init__(board)
        self._reader_functions = {
            "temp": self.read_temp_of_chip,
            "current_1v2": self.read_1v2_current_of_chip,
            "current_2v5": self.read_2v5_current_of_chip,
            "vadjn": self.read_vadjn,
            "vadjp": self.read_vadjp,
        }

    def read_1v2_current_of_chip(self):
        """Reads out the MCP3426 monitoring the current and converts to Ampere.

        Reads out the MCP3426 monitoring the current sense resistor, and converts
        it to a current in A.

        Returns:
            Current in Ampere.

        Raises:
            NotImplementedError if this functionality is not available for the current model.
        """
        rail = "current_1v2"

        return self._read_rail_current_of_chip(rail)

    def read_2v5_current_of_chip(self):
        """Reads out the MCP3426 monitoring the current and converts to Ampere.

        Reads out the MCP3426 monitoring the current sense resistor, and converts
        it to a current in A.

        Returns:
            Current in Ampere.

        Raises:
            NotImplementedError if this functionality is not available for the current model.
        """
        rail = "current_2v5"

        return self._read_rail_current_of_chip(rail)

    def _read_rail_current_of_chip(self, rail):
        voltage_mv = self._request_readout(rail)
        resistance_mo = self._get_v2a_resistance(rail)
        current_ma = self._convert_voltage_to_current(voltage_mv, resistance_mo=resistance_mo)
        logger.debug("Current Sense: %smA / %smV", current_ma, voltage_mv)
        return current_ma

    def _get_v2a_resistance(self, rail):
        """Return voltage to current resistor in mOhm."""
        try:
            return self.board.params["peripherals"][rail]["r_sense"] * 1000  # mOhm
        except KeyError:
            raise NotImplementedError(
                f"{rail} readout functionality has not been implemented for {self.board.model}"
            )

    def _convert_voltage_to_current(self, voltage_mv, resistance_mo):
        """Convert voltage to current by checking voltage drop over resistor."""
        current_ma = voltage_mv / resistance_mo
        return current_ma



"""IO module is for saving and loading from disk.

"""
import gzip
import os
import pickle
from collections import deque
from logging import getLogger
from pathlib import Path
from typing import Union

from naludaq.models import acq_converters
from naludaq.models.acquisition import Acquisition

from .io_manager import IOManager

LOGGER = getLogger(__name__)

disk_io = IOManager()


def save_file(filename: str, data):
    """Saves a python object as pickled data.

    This function should be upgraded to a proper factory.
    """
    if filename is None:
        raise TypeError("Supplied pathname is NoneType.")

    path, _ = os.path.split(filename)
    if not os.path.isdir(path):
        raise NotADirectoryError(f"Not a valid directory: {path}")

    try:
        sfile = gzip.GzipFile(filename, "w", compresslevel=4)
        pickle.dump(data, sfile, protocol=pickle.HIGHEST_PROTOCOL)
    except IOError as error_msg:
        LOGGER.error("File could not be created. %s", error_msg)
    except pickle.PicklingError as error_msg:
        LOGGER.error("Saving acquisition failed: %s", error_msg)


def load_file(filename: str):
    """Loads a python pickled file.

    This function should be upgraded to a factory.
    """
    if not os.path.isfile(filename):
        raise FileNotFoundError(f"No file found: {filename}")

    try:
        sfile = gzip.GzipFile(filename, "r")
        output = pickle.load(sfile)
    except EOFError as error_msg:
        LOGGER.error("Reached the end of the file, msg: %s", error_msg)
    except IOError as error_msg:
        LOGGER.error("File could not be loaded. %s", error_msg)
    except pickle.UnpicklingError as error_msg:
        LOGGER.error("Loading pedestals failed: %s", error_msg)
    else:
        return output


def load_pickle_acquisition(path: Union[str, Path]) -> Acquisition:
    """Loads an acquisition from a pickle file.

    Will upgrade files using the old acquisition format
    to the current version.

    Args:
        path (str, Path): the file on disk to load.

    Raises:
        FileNotFoundError if the file does not exist or is not a file.
        EOFError, UnpicklingError if the file is not a valid gzip/pickle file.
        IOError if there was a problem opening the file.

    Returns:
        The loaded acquisition
    """
    if not os.path.isfile(path):
        raise FileNotFoundError(f"No file found: {path}")

    try:
        with gzip.GzipFile(path, "rb") as f:
            output = pickle.load(f)
    except EOFError as error_msg:
        LOGGER.error("Reached the end of the file, msg: %s", error_msg)
        raise error_msg
    except IOError as error_msg:
        LOGGER.error("File could not be loaded. %s", error_msg)
        raise error_msg
    except pickle.UnpicklingError as error_msg:
        LOGGER.error("File could not be loaded. %s", error_msg)
        raise error_msg

    # Upgrade all acquisitions. Converting to the same version is cheap
    return acq_converters.upgrade_old_acquisition(output)


def save_pickle_acquisition(acq: Union[deque, Acquisition], path: Union[str, Path]):
    """Saves an acquisition or deque to a pickle file on the disk.

    Acquisitions are pickled as a dict so they can be opened elsewhere
    without naludaq installed.

    Args:
        acq (deque, Acquisition): the acquisition to save
        path (str, Path): the output file location

    Raises:
        TypeError if the given acquisition is not a deque or `Acquisition`.
        IOError if there was a problem opening the file.
    """
    if not isinstance(acq, (Acquisition, deque)):
        raise TypeError(f"Expected an Acquisition or deque, not {type(acq)}")
    try:
        with gzip.GzipFile(path, "wb", compresslevel=4) as f:
            if isinstance(acq, Acquisition):
                pickle.dump(acq.as_dict(), f, protocol=pickle.HIGHEST_PROTOCOL)
            elif isinstance(acq, deque):
                pickle.dump(acq, f, protocol=pickle.HIGHEST_PROTOCOL)
    except IOError as error_msg:
        LOGGER.error("File could not be loaded. %s", error_msg)
        raise error_msg
    except pickle.PicklingError as error_msg:
        LOGGER.error("File could not be saved. %s", error_msg)
        raise error_msg
